import pytest

from asyncpraw.models import RemovalReason, Subreddit
from asyncpraw.models.reddit.removal_reasons import SubredditRemovalReasons

from ... import UnitTest


class TestRemovalReason(UnitTest):
    def test_equality(self, reddit):
        reason1 = RemovalReason(
            reddit, subreddit=Subreddit(reddit, display_name="a"), id="x"
        )
        reason2 = RemovalReason(
            reddit, subreddit=Subreddit(reddit, display_name="a"), id="2"
        )
        reason3 = RemovalReason(
            reddit, subreddit=Subreddit(reddit, display_name="b"), id="1"
        )
        reason4 = RemovalReason(
            reddit, subreddit=Subreddit(reddit, display_name="A"), id="x"
        )
        reason5 = RemovalReason(
            reddit,
            subreddit=Subreddit(reddit, display_name="a"),
            reason_id="X",
        )
        assert reason1 == reason1
        assert reason1 == "x"
        assert reason2 == reason2
        assert reason3 == reason3
        assert reason1 != reason2
        assert reason1 != reason3
        assert reason1 == reason4
        assert reason1 != reason5

    def test_exception(self, reddit):
        with pytest.raises(ValueError):
            RemovalReason(reddit, subreddit=Subreddit(reddit, display_name="a"))
        with pytest.raises(ValueError):
            RemovalReason(
                reddit,
                subreddit=Subreddit(reddit, display_name="a"),
                id="test",
                _data={},
            )
        with pytest.raises(ValueError):
            RemovalReason(reddit, subreddit=Subreddit(reddit, display_name="a"), id="")
        with pytest.raises(ValueError):
            RemovalReason(
                reddit,
                subreddit=Subreddit(reddit, display_name="a"),
                reason_id="",
            )

    async def test__get(self, reddit):
        subreddit = Subreddit(reddit, display_name="a")
        removal_reason = await subreddit.mod.removal_reasons.get_reason(
            "a", fetch=False
        )
        assert isinstance(removal_reason, RemovalReason)

    def test_hash(self, reddit):
        reason1 = RemovalReason(
            reddit, subreddit=Subreddit(reddit, display_name="a"), id="x"
        )
        reason2 = RemovalReason(
            reddit, subreddit=Subreddit(reddit, display_name="a"), id="2"
        )
        reason3 = RemovalReason(
            reddit, subreddit=Subreddit(reddit, display_name="b"), id="1"
        )
        reason4 = RemovalReason(
            reddit, subreddit=Subreddit(reddit, display_name="A"), id="x"
        )
        reason5 = RemovalReason(
            reddit,
            subreddit=Subreddit(reddit, display_name="a"),
            reason_id="X",
        )
        assert hash(reason1) == hash(reason1)
        assert hash(reason2) == hash(reason2)
        assert hash(reason3) == hash(reason3)
        assert hash(reason1) != hash(reason2)
        assert hash(reason1) != hash(reason3)
        assert hash(reason1) == hash(reason4)
        assert hash(reason1) != hash(reason5)

    def test_repr(self, reddit):
        reason = RemovalReason(
            reddit, subreddit=Subreddit(reddit, display_name="a"), id="x"
        )
        assert repr(reason) == "RemovalReason(id='x')"

    def test_str(self, reddit):
        reason = RemovalReason(
            reddit, subreddit=Subreddit(reddit, display_name="a"), id="x"
        )
        assert str(reason) == "x"


class TestSubredditRemovalReasons(UnitTest):
    def test_repr(self, reddit):
        sr = SubredditRemovalReasons(subreddit=Subreddit(reddit, display_name="a"))
        assert repr(sr)  # assert it has some repr



"""Test asyncpraw.models.poll."""


import pytest

from asyncpraw.models import PollData, PollOption

from ... import IntegrationTest


class TestPollData(IntegrationTest):
    async def test_get_attrs(self):
        poll_id = "fo7p5b"
        with self.use_cassette():
            submission = await self.reddit.submission(poll_id)
            poll_data = submission.poll_data
            assert isinstance(poll_data, PollData)
            assert isinstance(poll_data.options, list)
            assert all(isinstance(option, PollOption) for option in poll_data.options)
            assert poll_data.user_selection is None  # not logged in

            with pytest.raises(KeyError):
                poll_data.option("badID")

    async def test_get_user_selection(self):
        poll_id = "hlzjkc"
        self.reddit.read_only = False
        with self.use_cassette():
            submission = await self.reddit.submission(poll_id)
            poll_data = submission.poll_data
            # For this test to pass, the authenticated user should have
            # voted in the poll specified.
            assert isinstance(poll_data.user_selection, PollOption)



"""Provides the code to load Async PRAW's configuration file ``praw.ini``."""
import configparser
import os
import sys
from threading import Lock
from typing import Optional

from .exceptions import ClientException


class _NotSet:
    def __bool__(self):
        return False

    __nonzero__ = __bool__

    def __str__(self):
        return "NotSet"


class Config:
    """A class containing the configuration for a Reddit site."""

    CONFIG = None
    CONFIG_NOT_SET = _NotSet()  # Represents a config value that is not set.
    LOCK = Lock()
    INTERPOLATION_LEVEL = {
        "basic": configparser.BasicInterpolation,
        "extended": configparser.ExtendedInterpolation,
    }

    @staticmethod
    def _config_boolean(item):
        if isinstance(item, bool):
            return item
        return item.lower() in {"1", "yes", "true", "on"}

    @classmethod
    def _load_config(cls, *, config_interpolation: Optional[str] = None):
        """Attempt to load settings from various praw.ini files."""
        if config_interpolation is not None:
            interpolator_class = cls.INTERPOLATION_LEVEL[config_interpolation]()
        else:
            interpolator_class = None
        config = configparser.ConfigParser(interpolation=interpolator_class)
        module_dir = os.path.dirname(sys.modules[__name__].__file__)
        if "APPDATA" in os.environ:  # Windows
            os_config_path = os.environ["APPDATA"]
        elif "XDG_CONFIG_HOME" in os.environ:  # Modern Linux
            os_config_path = os.environ["XDG_CONFIG_HOME"]
        elif "HOME" in os.environ:  # Legacy Linux
            os_config_path = os.path.join(os.environ["HOME"], ".config")
        else:
            os_config_path = None
        locations = [os.path.join(module_dir, "praw.ini"), "praw.ini"]
        if os_config_path is not None:
            locations.insert(1, os.path.join(os_config_path, "praw.ini"))
        config.read(locations)
        cls.CONFIG = config

    @property
    def short_url(self) -> str:
        """Return the short url.

        :raises: :class:`.ClientException` if it is not set.

        """
        if self._short_url is self.CONFIG_NOT_SET:
            raise ClientException("No short domain specified.")
        return self._short_url

    def __init__(
        self,
        site_name: str,
        config_interpolation: Optional[str] = None,
        **settings: str,
    ):
        """Initialize a :class:`.Config` instance."""
        with Config.LOCK:
            if Config.CONFIG is None:
                self._load_config(config_interpolation=config_interpolation)

        self._settings = settings
        self.custom = dict(Config.CONFIG.items(site_name), **settings)

        self.client_id = self.client_secret = self.oauth_url = None
        self.reddit_url = self.refresh_token = self.redirect_uri = None
        self.password = self.user_agent = self.username = None

        self._initialize_attributes()

    def _fetch(self, key):
        value = self.custom[key]
        del self.custom[key]
        return value

    def _fetch_default(self, key, *, default=None):
        if key not in self.custom:
            return default
        return self._fetch(key)

    def _fetch_or_not_set(self, key):
        if key in self._settings:  # Passed in values have the highest priority
            return self._fetch(key)

        env_value = os.getenv(f"praw_{key}")
        ini_value = self._fetch_default(key)  # Needed to remove from custom

        # Environment variables have higher priority than praw.ini settings
        return env_value or ini_value or self.CONFIG_NOT_SET

    def _initialize_attributes(self):
        self._short_url = self._fetch_default("short_url") or self.CONFIG_NOT_SET
        self.check_for_updates = self._config_boolean(
            self._fetch_or_not_set("check_for_updates")
        )
        self.warn_comment_sort = self._config_boolean(
            self._fetch_default("warn_comment_sort", default=True)
        )
        self.kinds = {
            x: self._fetch(f"{x}_kind")
            for x in [
                "comment",
                "message",
                "redditor",
                "submission",
                "subreddit",
                "trophy",
            ]
        }

        for attribute in (
            "client_id",
            "client_secret",
            "redirect_uri",
            "refresh_token",
            "password",
            "user_agent",
            "username",
        ):
            setattr(self, attribute, self._fetch_or_not_set(attribute))

        for required_attribute in (
            "oauth_url",
            "ratelimit_seconds",
            "reddit_url",
            "timeout",
        ):
            setattr(self, required_attribute, self._fetch(required_attribute))

        for attribute, conversion in {
            "ratelimit_seconds": int,
            "timeout": int,
        }.items():
            try:
                setattr(self, attribute, conversion(getattr(self, attribute)))
            except ValueError:
                raise ValueError(
                    f"An incorrect config type was given for option {attribute}. The"
                    f" expected type is {conversion.__name__}, but the given value is"
                    f" {getattr(self, attribute)}."
                )



import pytest

from asyncpraw.models import Emoji, Subreddit
from asyncpraw.models.reddit.emoji import SubredditEmoji

from ... import UnitTest


class TestEmoji(UnitTest):
    def test_equality(self, reddit):
        emoji1 = Emoji(reddit, subreddit=Subreddit(reddit, "a"), name="x")
        emoji2 = Emoji(reddit, subreddit=Subreddit(reddit, "a"), name="2")
        emoji3 = Emoji(reddit, subreddit=Subreddit(reddit, "b"), name="1")
        emoji4 = Emoji(reddit, subreddit=Subreddit(reddit, "A"), name="x")
        emoji5 = Emoji(reddit, subreddit=Subreddit(reddit, "a"), name="X")
        emoji6 = Emoji(reddit, subreddit=Subreddit(reddit, "b"), name="x")
        assert emoji1 == emoji1
        assert emoji1 == "x"
        assert emoji2 == emoji2
        assert emoji3 == emoji3
        assert emoji1 != emoji2
        assert emoji1 != emoji3
        assert emoji1 == emoji4
        assert emoji1 != emoji5
        assert emoji1 != emoji6
        assert emoji1 != 5

    async def test__get(self, reddit):
        subreddit = Subreddit(reddit, display_name="a")
        emoji = await subreddit.emoji.get_emoji("a", fetch=False)
        assert isinstance(emoji, Emoji)

    def test_hash(self, reddit):
        emoji1 = Emoji(reddit, subreddit=Subreddit(reddit, "a"), name="x")
        emoji2 = Emoji(reddit, subreddit=Subreddit(reddit, "a"), name="2")
        emoji3 = Emoji(reddit, subreddit=Subreddit(reddit, "b"), name="1")
        emoji4 = Emoji(reddit, subreddit=Subreddit(reddit, "A"), name="x")
        emoji5 = Emoji(reddit, subreddit=Subreddit(reddit, "a"), name="X")
        emoji6 = Emoji(reddit, subreddit=Subreddit(reddit, "b"), name="x")
        assert hash(emoji1) == hash(emoji1)
        assert hash(emoji2) == hash(emoji2)
        assert hash(emoji3) == hash(emoji3)
        assert hash(emoji1) != hash(emoji2)
        assert hash(emoji1) != hash(emoji3)
        assert hash(emoji1) == hash(emoji4)
        assert hash(emoji1) != hash(emoji5)
        assert hash(emoji1) != hash(emoji6)

    def test_repr(self, reddit):
        emoji = Emoji(reddit, subreddit=Subreddit(reddit, "a"), name="x")
        assert repr(emoji) == "Emoji(name='x')"

    def test_str(self, reddit):
        emoji = Emoji(reddit, subreddit=Subreddit(reddit, "a"), name="x")
        assert str(emoji) == "x"

    async def test_update(self, reddit):
        emoji = Emoji(reddit, subreddit=Subreddit(reddit, "a"), name="x")
        with pytest.raises(TypeError) as excinfo:
            await emoji.update()
        assert str(excinfo.value) == "At least one attribute must be provided"


class TestSubredditEmoji(UnitTest):
    def test_repr(self, reddit):
        se = SubredditEmoji(subreddit=Subreddit(reddit, "a"))
        assert repr(se)  # assert it has some repr



"""Test asyncpraw.models.listing.generator."""
from ... import IntegrationTest


class TestListingGenerator(IntegrationTest):
    async def test_exhaust_items_with_before(self):
        with self.use_cassette():
            spez = await self.reddit.redditor("spez")
            submissions = await self.async_list(
                spez.top(limit=None, params={"before": "3cxedn"})
            )
        assert len(submissions) > 100

    async def test_exhaust_items(self):
        with self.use_cassette():
            spez = await self.reddit.redditor("spez")
            submissions = await self.async_list(spez.top(limit=None))
        assert len(submissions) > 100

    async def test_no_items(self):
        with self.use_cassette():
            spez = await self.reddit.redditor("spez")
            submissions = await self.async_list(spez.top(time_filter="hour"))
        assert len(submissions) == 0



"""Provide the RedditBase class."""
from typing import TYPE_CHECKING, Any, Dict, Optional, Union
from urllib.parse import urlparse

from ...exceptions import InvalidURL
from ..base import AsyncPRAWBase

if TYPE_CHECKING:  # pragma: no cover
    import asyncpraw


class RedditBase(AsyncPRAWBase):
    """Base class that represents actual Reddit objects."""

    @staticmethod
    def _url_parts(url):
        parsed = urlparse(url)
        if not parsed.netloc:
            raise InvalidURL(url)
        return parsed.path.rstrip("/").split("/")

    def __eq__(self, other: Union[Any, str]) -> bool:
        """Return whether the other instance equals the current."""
        if isinstance(other, str):
            return other.lower() == str(self).lower()
        return (
            isinstance(other, self.__class__)
            and str(self).lower() == str(other).lower()
        )

    def __getattr__(self, attribute: str) -> Any:
        """Return the value of ``attribute``."""
        if not attribute.startswith("_") and not self._fetched:
            raise AttributeError(
                "{0!r} object has no attribute {1!r}. {0!r} object has not been"
                " fetched, did you forget to execute '.load()'?".format(
                    self.__class__.__name__, attribute
                )
            )
        raise AttributeError(
            f"{self.__class__.__name__!r} object has no attribute {attribute!r}"
        )

    def __hash__(self) -> int:
        """Return the hash of the current instance."""
        return hash(self.__class__.__name__) ^ hash(str(self).lower())

    def __init__(
        self,
        reddit: "asyncpraw.Reddit",
        _data: Optional[Dict[str, Any]],
        _extra_attribute_to_check: Optional[str] = None,
        _fetched: bool = False,
        _str_field: bool = True,
    ):
        """Initialize a :class:`.RedditBase` instance.

        :param reddit: An instance of :class:`.Reddit`.

        """
        super().__init__(reddit, _data=_data)
        self._fetched = _fetched
        if _str_field and self.STR_FIELD not in self.__dict__:
            if (
                _extra_attribute_to_check is not None
                and _extra_attribute_to_check in self.__dict__
            ):
                return
            raise ValueError(
                f"An invalid value was specified for {self.STR_FIELD}. Check that the "
                f"argument for the {self.STR_FIELD} parameter is not empty."
            )

    def __repr__(self) -> str:
        """Return an object initialization representation of the instance."""
        return f"{self.__class__.__name__}({self.STR_FIELD}={str(self)!r})"

    def __str__(self) -> str:
        """Return a string representation of the instance."""
        return getattr(self, self.STR_FIELD)

    def __ne__(self, other: Any) -> bool:
        """Return whether the other instance differs from the current."""
        return not self == other

    async def _fetch(self):  # pragma: no cover
        self._fetched = True

    def _reset_attributes(self, *attributes):
        for attribute in attributes:
            if attribute in self.__dict__:
                del self.__dict__[attribute]
        self._fetched = False

    async def load(self):
        """Re-fetches the object.

        This is used to explicitly fetch or re-fetch the object from reddit. This method
        can be used on any :class:`.RedditBase` object.

        .. code-block:: python

            await reddit_base_object.load()

        """
        await self._fetch()



"""Test asyncpraw.util.refresh_token_manager."""
import sys
from tempfile import NamedTemporaryFile

import aiofiles
import pytest

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.util.token_manager import (
    BaseTokenManager,
    FileTokenManager,
    SQLiteTokenManager,
)

from .. import UnitTest
from ..test_reddit import DummyTokenManager


class DummyAuthorizer:
    def __init__(self, refresh_token):
        self.refresh_token = refresh_token


class TestBaseTokenManager(UnitTest):
    def test_init_base_fail(self):
        with pytest.raises(TypeError):
            BaseTokenManager()

    def test_reddit(self):
        manager = DummyTokenManager()
        manager.reddit = "dummy"
        assert manager.reddit == "dummy"

    def test_reddit__must_only_be_set_once(self):
        manager = DummyTokenManager()
        manager.reddit = "dummy"
        with pytest.raises(RuntimeError) as excinfo:
            manager.reddit = None
        assert (
            str(excinfo.value)
            == "'reddit' can only be set once and is done automatically"
        )


class TestFileTokenManager(UnitTest):
    def setup(self):
        aiofiles.threadpool.wrap.register(mock.MagicMock)(
            lambda *args, **kwargs: aiofiles.threadpool.AsyncBufferedIOBase(
                *args, **kwargs
            )
        )

    async def test_post_refresh_token_callback__writes_to_file(self):
        authorizer = DummyAuthorizer("token_value")
        manager = FileTokenManager("mock/dummy_path")
        mock_file = mock.MagicMock()

        with mock.patch(
            "aiofiles.threadpool.sync_open", return_value=mock_file
        ) as mock_open:
            await manager.post_refresh_callback(authorizer)

            assert authorizer.refresh_token == "token_value"
            mock_open.assert_called_once_with(
                "mock/dummy_path",
                mode="w",
                buffering=-1,
                encoding=None,
                errors=None,
                newline=None,
                closefd=True,
                opener=None,
            )
            mock_open().write.assert_called_once_with("token_value")

    async def test_pre_refresh_token_callback__reads_from_file(self):
        authorizer = DummyAuthorizer(None)
        manager = FileTokenManager("mock/dummy_path")
        mock_file = mock.MagicMock()
        mock_file.read = mock.MagicMock(return_value="token_value\n")

        with mock.patch(
            "aiofiles.threadpool.sync_open", return_value=mock_file
        ) as mock_open:
            await manager.pre_refresh_callback(authorizer)

            assert authorizer.refresh_token == "token_value"
            mock_open.assert_called_once_with(
                "mock/dummy_path",
                mode="r",
                buffering=-1,
                encoding=None,
                errors=None,
                newline=None,
                closefd=True,
                opener=None,
            )


class TestSQLiteTokenManager(UnitTest):
    def setup(self):
        self.manager = SQLiteTokenManager(database=":memory:", key="dummy_key")

    async def test_is_registered(self):
        assert not await self.manager.is_registered()
        await self.manager.close()

    @pytest.mark.skipif(
        sys.platform.startswith("win"), reason="this test fails on windows"
    )
    async def test_multiple_instances(self):
        with NamedTemporaryFile() as fp:
            manager1 = SQLiteTokenManager(database=fp.name, key="dummy_key1")
            manager2 = SQLiteTokenManager(database=fp.name, key="dummy_key1")
            manager3 = SQLiteTokenManager(database=fp.name, key="dummy_key2")

            await manager1.register("dummy_value1")
            assert await manager2.is_registered()
            assert not await manager3.is_registered()
            await manager1.close()
            await manager2.close()
            await manager3.close()

    async def test_post_refresh_token_callback__sets_value(self):
        authorizer = DummyAuthorizer("dummy_value")

        await self.manager.post_refresh_callback(authorizer)
        assert authorizer.refresh_token is None
        assert await self.manager._get() == "dummy_value"
        await self.manager.close()

    async def test_post_refresh_token_callback__updates_value(self):
        authorizer = DummyAuthorizer("dummy_value_updated")
        await self.manager.register("dummy_value")

        await self.manager.post_refresh_callback(authorizer)
        assert authorizer.refresh_token is None
        assert await self.manager._get() == "dummy_value_updated"
        await self.manager.close()

    async def test_pre_refresh_token_callback(self):
        authorizer = DummyAuthorizer(None)
        await self.manager.register("dummy_value")

        await self.manager.pre_refresh_callback(authorizer)
        assert authorizer.refresh_token == "dummy_value"
        await self.manager.close()

    async def test_pre_refresh_token_callback__raises_key_error(self):
        authorizer = DummyAuthorizer(None)

        with pytest.raises(KeyError):
            await self.manager.pre_refresh_callback(authorizer)
        await self.manager.close()

    async def test_register(self):
        assert await self.manager.register("dummy_value")
        assert await self.manager.is_registered()
        assert not await self.manager.register("dummy_value2")
        assert await self.manager._get() == "dummy_value"
        await self.manager.close()



"""Provide classes related to widgets."""

import os.path
from json import JSONEncoder, dumps
from typing import TYPE_CHECKING, Any, Dict, List, Union

from ...const import API_PATH
from ...util import _deprecate_args
from ...util.cache import cachedproperty
from ..base import AsyncPRAWBase
from ..list.base import BaseList

if TYPE_CHECKING:  # pragma: no cover
    import asyncpraw

WidgetType = Union[
    "asyncpraw.models.ButtonWidget",
    "asyncpraw.models.Calendar",
    "asyncpraw.models.CommunityList",
    "asyncpraw.models.CustomWidget",
    "asyncpraw.models.IDCard",
    "asyncpraw.models.ImageWidget",
    "asyncpraw.models.Menu",
    "asyncpraw.models.ModeratorsWidget",
    "asyncpraw.models.PostFlairWidget",
    "asyncpraw.models.RulesWidget",
    "asyncpraw.models.TextArea",
    "asyncpraw.models.Widget",
]


class Button(AsyncPRAWBase):
    """Class to represent a single button inside a :class:`.ButtonWidget`.

    .. include:: ../../typical_attributes.rst

    ============== =====================================================================
    Attribute      Description
    ============== =====================================================================
    ``color``      The hex color used to outline the button.
    ``fillColor``  The hex color for the background of the button.
    ``height``     Image height. Only present on image buttons.
    ``hoverState`` A ``dict`` describing the state of the button when hovered over.
                   Optional.
    ``kind``       Either ``"text"`` or ``"image"``.
    ``linkUrl``    A link that can be visited by clicking the button. Only present on
                   image buttons.
    ``text``       The text displayed on the button.
    ``textColor``  The hex color for the text of the button.
    ``url``        - If the button is a text button, a link that can be visited by
                     clicking the button.
                   - If the button is an image button, the URL of a Reddit-hosted image.
    ``width``      Image width. Only present on image buttons.
    ============== =====================================================================

    """


class CalendarConfiguration(AsyncPRAWBase):
    """Class to represent the configuration of a :class:`.Calendar`.

    .. include:: ../../typical_attributes.rst

    =================== ================================================
    Attribute           Description
    =================== ================================================
    ``numEvents``       The number of events to display on the calendar.
    ``showDate``        Whether to show the dates of events.
    ``showDescription`` Whether to show the descriptions of events.
    ``showLocation``    Whether to show the locations of events.
    ``showTime``        Whether to show the times of events.
    ``showTitle``       Whether to show the titles of events.
    =================== ================================================

    """


class Hover(AsyncPRAWBase):
    """Class to represent the hover data for a :class:`.ButtonWidget`.

    These values will take effect when the button is hovered over (the user moves their
    cursor so it's on top of the button).

    .. include:: ../../typical_attributes.rst

    ============= =====================================================================
    Attribute     Description
    ============= =====================================================================
    ``color``     The hex color used to outline the button.
    ``fillColor`` The hex color for the background of the button.
    ``textColor`` The hex color for the text of the button.
    ``height``    Image height. Only present on image buttons.
    ``kind``      Either ``text`` or ``image``.
    ``text``      The text displayed on the button.
    ``url``       - If the button is a text button, a link that can be visited by
                    clicking the button.
                  - If the button is an image button, the URL of a Reddit-hosted image.
    ``width``     Image width. Only present on image buttons.
    ============= =====================================================================

    """


class Image(AsyncPRAWBase):
    """Class to represent an image that's part of a :class:`.ImageWidget`.

    .. include:: ../../typical_attributes.rst

    =========== =================================================
    Attribute   Description
    =========== =================================================
    ``height``  Image height.
    ``linkUrl`` A link that can be visited by clicking the image.
    ``url``     The URL of the (Reddit-hosted) image.
    ``width``   Image width.
    =========== =================================================

    """


class ImageData(AsyncPRAWBase):
    """Class for image data that's part of a :class:`.CustomWidget`.

    .. include:: ../../typical_attributes.rst

    ========== =========================================
    Attribute  Description
    ========== =========================================
    ``height`` The image height.
    ``name``   The image name.
    ``url``    The URL of the image on Reddit's servers.
    ``width``  The image width.
    ========== =========================================

    """


class MenuLink(AsyncPRAWBase):
    """Class to represent a single link inside a :class:`.Menu` or :class:`.Submenu`.

    .. include:: ../../typical_attributes.rst

    ========= ====================================
    Attribute Description
    ========= ====================================
    ``text``  The text of the menu link.
    ``url``   The URL that the menu item links to.
    ========= ====================================

    """


class Styles(AsyncPRAWBase):
    """Class to represent the style information of a widget.

    .. include:: ../../typical_attributes.rst

    =================== ========================================================
    Attribute           Description
    =================== ========================================================
    ``backgroundColor`` The background color of a widget, given as a hexadecimal
                        (``0x######``).
    ``headerColor``     The header color of a widget, given as a hexadecimal
                        (``0x######``).
    =================== ========================================================

    """


class Submenu(BaseList):
    r"""Class to represent a submenu of links inside a :class:`.Menu`.

    .. include:: ../../typical_attributes.rst

    ============ ======================================================================
    Attribute    Description
    ============ ======================================================================
    ``children`` A list of the :class:`.MenuLink`\ s in this submenu. Can be iterated
                 over by iterating over the :class:`.Submenu` (e.g., ``for menu_link in
                 submenu``).
    ``text``     The name of the submenu.
    ============ ======================================================================

    """

    CHILD_ATTRIBUTE = "children"


class SubredditWidgets(AsyncPRAWBase):
    """Class to represent a :class:`.Subreddit`'s widgets.

    Create an instance like so:

    .. code-block:: python

        subreddit = await reddit.subreddit("test")
        widgets = subreddit.widgets

    Data will be lazy-loaded. By default, Async PRAW will not request progressively
    loading images from Reddit. To enable this, instantiate a :class:`.SubredditWidgets`
    object via :meth:`~.Subreddit.widgets`, then set the attribute
    ``progressive_images`` to ``True`` before performing any action that would result in
    a network request.

    .. code-block:: python

        subreddit = await reddit.subreddit("test")
        widgets = subreddit.widgets
        widgets.progressive_images = True
        async for widget in widgets.sidebar():
            # do something
            ...

    Access a :class:`.Subreddit`'s widgets with the following attributes:

    .. code-block:: python

        print(await widgets.id_card())
        print(await widgets.moderators_widget())
        print([widget async for widget in widgets.sidebar()])
        print([widget async for widget in widgets.topbar()])

    The attribute :attr:`.id_card` contains the :class:`.Subreddit`'s ID card, which
    displays information like the number of subscribers.

    The attribute :attr:`.moderators_widget` contains the :class:`.Subreddit`'s
    moderators widget, which lists the moderators of the subreddit.

    The attribute :attr:`.sidebar` contains a list of widgets which make up the sidebar
    of the subreddit.

    The attribute :attr:`.topbar` contains a list of widgets which make up the top bar
    of the subreddit.

    To edit a :class:`.Subreddit`'s widgets, use :attr:`~.SubredditWidgets.mod`. For
    example:

    .. code-block:: python

        await widgets.mod.add_text_area(
            short_name="My title",
            text="**bold text**",
            styles={"backgroundColor": "#FFFF66", "headerColor": "#3333EE"},
        )

    For more information, see :class:`.SubredditWidgetsModeration`.

    To edit a particular widget, use ``.mod`` on the widget. For example:

    .. code-block:: python

        async for widget in widgets.sidebar():
            await widget.mod.update(shortName="Exciting new name")

    For more information, see :class:`.WidgetModeration`.

    **Currently available widgets**:

    - :class:`.ButtonWidget`
    - :class:`.Calendar`
    - :class:`.CommunityList`
    - :class:`.CustomWidget`
    - :class:`.IDCard`
    - :class:`.ImageWidget`
    - :class:`.Menu`
    - :class:`.ModeratorsWidget`
    - :class:`.PostFlairWidget`
    - :class:`.RulesWidget`
    - :class:`.TextArea`

    """

    async def id_card(self) -> "asyncpraw.models.IDCard":
        """Get this :class:`.Subreddit`'s :class:`.IDCard` widget."""
        items = await self.items()
        return items[self.layout["idCardWidget"]]

    async def items(self) -> Dict[str, "asyncpraw.models.Widget"]:
        """Get this :class:`.Subreddit`'s widgets as a dict from ID to widget."""
        if self._items is None:
            if not self._raw_items:
                await self._fetch()
            self._items = {}
            for item_name, data in self._raw_items.items():
                data["subreddit"] = self.subreddit
                self._items[item_name] = self._reddit._objector.objectify(data)
        return self._items

    @cachedproperty
    def mod(self) -> "asyncpraw.models.SubredditWidgetsModeration":
        """Get an instance of :class:`.SubredditWidgetsModeration`.

        .. note::

            Using any of the methods of :class:`.SubredditWidgetsModeration` will likely
            result in the data of this :class:`.SubredditWidgets` being outdated. To
            re-sync, call :meth:`.refresh`.

        """
        return SubredditWidgetsModeration(self.subreddit, self._reddit)

    async def moderators_widget(self) -> "asyncpraw.models.ModeratorsWidget":
        """Get this :class:`.Subreddit`'s :class:`.ModeratorsWidget`."""
        items = await self.items()
        return items[self.layout["moderatorWidget"]]

    async def sidebar(self) -> List["asyncpraw.models.Widget"]:
        r"""Get a list of :class:`.Widget`\ s that make up the sidebar."""
        items = await self.items()
        for widget in self.layout["sidebar"]["order"]:
            yield items[widget]

    async def topbar(self) -> List["asyncpraw.models.Menu"]:
        r"""Get a list of :class:`.Widget`\ s that make up the top bar."""
        items = await self.items()
        for widget in self.layout["topbar"]["order"]:
            yield items[widget]

    async def refresh(self):
        """Refresh the :class:`.Subreddit`'s widgets.

        By default, Async PRAW will not request progressively loading images from
        Reddit. To enable this, set the attribute ``progressive_images`` to ``True``
        prior to calling ``refresh()``.

        .. code-block:: python

            subreddit = await reddit.subreddit("test")
            widgets = subreddit.widgets
            widgets.progressive_images = True
            await widgets.refresh()

        """
        await self._fetch()

    def __getattr__(self, attr: str) -> Any:
        """Return the value of ``attr``."""
        if not attr.startswith("_") and not self._fetched:
            raise AttributeError(
                f"{self.__class__.__name__!r} object has no attribute {attr!r}, did you"
                " forget to run '.refresh()'?"
            )
        raise AttributeError(  # pragma: no cover; I have no idea how to cover this
            f"{self.__class__.__name__!r} object has no attribute {attr!r}, did you"
            " forget to run '.refresh()'?"
        )

    def __init__(self, subreddit: "asyncpraw.models.Subreddit"):
        """Initialize a :class:`.SubredditWidgets` instance.

        :param subreddit: The :class:`.Subreddit` the widgets belong to.

        """
        self._raw_items = None
        self._items = None
        self._fetched = False
        self.subreddit = subreddit
        self.progressive_images = False

        super().__init__(subreddit._reddit, {})

    def __repr__(self) -> str:
        """Return an object initialization representation of the instance."""
        return f"SubredditWidgets(subreddit={self.subreddit!r})"

    async def _fetch(self):
        data = await self._reddit.get(
            API_PATH["widgets"].format(subreddit=self.subreddit),
            params={"progressive_images": self.progressive_images},
        )

        self._raw_items = data.pop("items")
        self._items = None
        super().__init__(self.subreddit._reddit, data)

        self._fetched = True


class SubredditWidgetsModeration:
    """Class for moderating a :class:`.Subreddit`'s widgets.

    Get an instance of this class from :attr:`.SubredditWidgets.mod`.

    Example usage:

    .. code-block:: python

        styles = {"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}
        subreddit = await reddit.subreddit("test")
        await subreddit.widgets.mod.add_text_area(
            short_name="My title", text="**bold text**", styles=styles
        )

    .. note::

        To use this class's methods, the authenticated user must be a moderator with
        appropriate permissions.

    """

    def __init__(
        self, subreddit: "asyncpraw.models.Subreddit", reddit: "asyncpraw.Reddit"
    ):
        """Initialize a :class:`.SubredditWidgetsModeration` instance."""
        self._subreddit = subreddit
        self._reddit = reddit

    async def _create_widget(self, payload: Dict[str, Any]) -> WidgetType:
        path = API_PATH["widget_create"].format(subreddit=self._subreddit)
        widget = await self._reddit.post(
            path, data={"json": dumps(payload, cls=WidgetEncoder)}
        )
        widget.subreddit = self._subreddit
        return widget

    @_deprecate_args("short_name", "description", "buttons", "styles")
    async def add_button_widget(
        self,
        *,
        buttons: List[
            Dict[str, Union[Dict[str, str], str, int, Dict[str, Union[str, int]]]]
        ],
        description: str,
        short_name: str,
        styles: Dict[str, str],
        **other_settings,
    ) -> "asyncpraw.models.ButtonWidget":
        """Add and return a :class:`.ButtonWidget`.

        :param buttons: A list of dictionaries describing buttons, as specified in
            `Reddit docs`_. As of this writing, the format is:

            Each button is either a text button or an image button. A text button looks
            like this:

            .. code-block:: text

                {
                    "kind": "text",
                    "text": a string no longer than 30 characters,
                    "url": a valid URL,
                    "color": a 6-digit rgb hex color, e.g., `#AABBCC`,
                    "textColor": a 6-digit rgb hex color, e.g., `#AABBCC`,
                    "fillColor": a 6-digit rgb hex color, e.g., `#AABBCC`,
                    "hoverState": {...}
                }

            An image button looks like this:

            .. code-block:: text

                {
                    "kind": "image",
                    "text": a string no longer than 30 characters,
                    "linkUrl": a valid URL,
                    "url": a valid URL of a Reddit-hosted image,
                    "height": an integer,
                    "width": an integer,
                    "hoverState": {...}
                }

            Both types of buttons have the field ``hoverState``. The field does not have
            to be included (it is optional). If it is included, it can be one of two
            types: ``"text"`` or ``"image"``. A text ``hoverState`` looks like this:

            .. code-block:: text

                {
                    "kind": "text",
                    "text": a string no longer than 30 characters,
                    "color": a 6-digit rgb hex color, e.g., `#AABBCC`,
                    "textColor": a 6-digit rgb hex color, e.g., `#AABBCC`,
                    "fillColor": a 6-digit rgb hex color, e.g., `#AABBCC`
                }

            An image ``hoverState`` looks like this:

            .. code-block:: text

                {
                    "kind": "image",
                    "url": a valid URL of a Reddit-hosted image,
                    "height": an integer,
                    "width": an integer
                }

            .. note::

                The method :meth:`.upload_image` can be used to upload images to Reddit
                for a ``url`` field that holds a Reddit-hosted image.

            .. note::

                An image ``hoverState`` may be paired with a text widget, and a text
                ``hoverState`` may be paired with an image widget.

        :param description: Markdown text to describe the widget.
        :param short_name: A name for the widget, no longer than 30 characters.
        :param styles: A dictionary with keys ``"backgroundColor"`` and
            ``"headerColor"``, and values of hex colors. For example,
            ``{"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}``.

        :returns: The created :class:`.ButtonWidget`.

        .. _reddit docs: https://www.reddit.com/dev/api#POST_api_widget

        Example usage:

        .. code-block:: python

            subreddit = await reddit.subreddit("test")
            widget_moderation = subreddit.widgets.mod
            my_image = await widget_moderation.upload_image("/path/to/pic.jpg")
            buttons = [
                {
                    "kind": "text",
                    "text": "View source",
                    "url": "https://github.com/asyncpraw-dev/asyncpraw",
                    "color": "#FF0000",
                    "textColor": "#00FF00",
                    "fillColor": "#0000FF",
                    "hoverState": {
                        "kind": "text",
                        "text": "ecruos weiV",
                        "color": "#FFFFFF",
                        "textColor": "#000000",
                        "fillColor": "#0000FF",
                    },
                },
                {
                    "kind": "image",
                    "text": "View documentation",
                    "linkUrl": "https://asyncpraw.readthedocs.io",
                    "url": my_image,
                    "height": 200,
                    "width": 200,
                    "hoverState": {"kind": "image", "url": my_image, "height": 200, "width": 200},
                },
            ]
            styles = {"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}
            new_widget = await widget_moderation.add_button_widget(
                short_name="Things to click",
                description="Click some of these *cool* links!",
                buttons=buttons,
                styles=styles,
            )

        """
        button_widget = {
            "buttons": buttons,
            "description": description,
            "kind": "button",
            "shortName": short_name,
            "styles": styles,
        }
        button_widget.update(other_settings)
        return await self._create_widget(button_widget)

    @_deprecate_args(
        "short_name", "google_calendar_id", "requires_sync", "configuration", "styles"
    )
    async def add_calendar(
        self,
        *,
        configuration: Dict[str, Union[bool, int]],
        google_calendar_id: str,
        requires_sync: bool,
        short_name: str,
        styles: Dict[str, str],
        **other_settings,
    ) -> "asyncpraw.models.Calendar":
        """Add and return a :class:`.Calendar` widget.

        :param configuration: A dictionary as specified in `Reddit docs`_. For example:

            .. code-block:: python

                {
                    "numEvents": 10,
                    "showDate": True,
                    "showDescription": False,
                    "showLocation": False,
                    "showTime": True,
                    "showTitle": True,
                }

        :param google_calendar_id: An email-style calendar ID. To share a Google
            Calendar, make it public, then find the "Calendar ID".
        :param requires_sync: Whether the calendar needs synchronization.
        :param short_name: A name for the widget, no longer than 30 characters.
        :param styles: A dictionary with keys ``"backgroundColor"`` and
            ``"headerColor"``, and values of hex colors. For example,
            ``{"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}``.

        :returns: The created :class:`.Calendar`.

        .. _reddit docs: https://www.reddit.com/dev/api#POST_api_widget

        Example usage:

        .. code-block:: python

            subreddit = await reddit.subreddit("test")
            widget_moderation = subreddit.widgets.mod
            styles = {"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}
            config = {
                "numEvents": 10,
                "showDate": True,
                "showDescription": False,
                "showLocation": False,
                "showTime": True,
                "showTitle": True,
            }
            calendar_id = "y6nm89jy427drk8l71w75w9wjn@group.calendar.google.com"
            new_widget = await widget_moderation.add_calendar(
                short_name="Upcoming Events",
                google_calendar_id=calendar_id,
                requires_sync=True,
                configuration=config,
                styles=styles,
            )

        """
        calendar = {
            "shortName": short_name,
            "googleCalendarId": google_calendar_id,
            "requiresSync": requires_sync,
            "configuration": configuration,
            "styles": styles,
            "kind": "calendar",
        }
        calendar.update(other_settings)
        return await self._create_widget(calendar)

    @_deprecate_args("short_name", "data", "styles", "description")
    async def add_community_list(
        self,
        *,
        data: List[Union[str, "asyncpraw.models.Subreddit"]],
        description: str = "",
        short_name: str,
        styles: Dict[str, str],
        **other_settings,
    ) -> "asyncpraw.models.CommunityList":
        """Add and return a :class:`.CommunityList` widget.

        :param data: A list of subreddits. Subreddits can be represented as ``str`` or
            as :class:`.Subreddit`. These types may be mixed within the list.
        :param description: A string containing Markdown (default: ``""``).
        :param short_name: A name for the widget, no longer than 30 characters.
        :param styles: A dictionary with keys ``"backgroundColor"`` and
            ``"headerColor"``, and values of hex colors. For example,
            ``{"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}``.

        :returns: The created :class:`.CommunityList`.

        Example usage:

        .. code-block:: python

            subreddit = await reddit.subreddit("test")
            widget_moderation = subreddit.widgets.mod
            styles = {"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}
            new_subreddit = await reddit.subreddit("test")
            subreddits = ["learnpython", new_subreddit]
            new_widget = await widget_moderation.add_community_list(
                short_name="My fav subs", data=subreddits, styles=styles, description="description"
            )

        """
        community_list = {
            "data": data,
            "kind": "community-list",
            "shortName": short_name,
            "styles": styles,
            "description": description,
        }
        community_list.update(other_settings)
        return await self._create_widget(community_list)

    @_deprecate_args("short_name", "text", "css", "height", "image_data", "styles")
    async def add_custom_widget(
        self,
        *,
        css: str,
        height: int,
        image_data: List[Dict[str, Union[str, int]]],
        short_name: str,
        styles: Dict[str, str],
        text: str,
        **other_settings,
    ) -> "asyncpraw.models.CustomWidget":
        """Add and return a :class:`.CustomWidget`.

        :param css: The CSS for the widget, no longer than 100000 characters.

            .. note::

                As of this writing, Reddit will not accept empty CSS. If you wish to
                create a custom widget without CSS, consider using ``"/**/"`` (an empty
                comment) as your CSS.

        :param height: The height of the widget, between 50 and 500.
        :param image_data: A list of dictionaries as specified in `Reddit docs`_. Each
            dictionary represents an image and has the key ``"url"`` which maps to the
            URL of an image hosted on Reddit's servers. Images should be uploaded using
            :meth:`.upload_image`.

            For example:

            .. code-block:: python

                [
                    {
                        "url": "https://some.link",  # from upload_image()
                        "width": 600,
                        "height": 450,
                        "name": "logo",
                    },
                    {
                        "url": "https://other.link",  # from upload_image()
                        "width": 450,
                        "height": 600,
                        "name": "icon",
                    },
                ]

        :param short_name: A name for the widget, no longer than 30 characters.
        :param styles: A dictionary with keys ``"backgroundColor"`` and
            ``"headerColor"``, and values of hex colors. For example,
            ``{"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}``.
        :param text: The Markdown text displayed in the widget.

        :returns: The created :class:`.CustomWidget`.

        .. _reddit docs: https://www.reddit.com/dev/api#POST_api_widget

        Example usage:

        .. code-block:: python

            subreddit = await reddit.subreddit("test")
            widget_moderation = subreddit.widgets.mod
            image_paths = ["/path/to/image1.jpg", "/path/to/image2.png"]
            image_urls = [widget_moderation.upload_image(img_path) for img_path in image_paths]
            image_data = [
                {"width": 600, "height": 450, "name": "logo", "url": image_urls[0]},
                {"width": 450, "height": 600, "name": "icon", "url": image_urls[1]},
            ]
            styles = {"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}
            new_widget = await widget_moderation.add_custom_widget(
                image_short_name="My widget",
                text="# Hello world!",
                css="/**/",
                height=200,
                image_data=image_data,
                styles=styles,
            )

        """
        custom_widget = {
            "css": css,
            "height": height,
            "imageData": image_data,
            "kind": "custom",
            "shortName": short_name,
            "styles": styles,
            "text": text,
        }
        custom_widget.update(other_settings)
        return await self._create_widget(custom_widget)

    @_deprecate_args("short_name", "data", "styles")
    async def add_image_widget(
        self,
        *,
        data: List[Dict[str, Union[str, int]]],
        short_name: str,
        styles: Dict[str, str],
        **other_settings,
    ) -> "asyncpraw.models.ImageWidget":
        """Add and return an :class:`.ImageWidget`.

        :param data: A list of dictionaries as specified in `Reddit docs`_. Each
            dictionary has the key ``"url"`` which maps to the URL of an image hosted on
            Reddit's servers. Images should be uploaded using :meth:`.upload_image`.

            For example:

            .. code-block:: python

                [
                    {
                        "url": "https://some.link",  # from upload_image()
                        "width": 600,
                        "height": 450,
                        "linkUrl": "https://github.com/asyncpraw-dev/asyncpraw",
                    },
                    {
                        "url": "https://other.link",  # from upload_image()
                        "width": 450,
                        "height": 600,
                        "linkUrl": "https://asyncpraw.readthedocs.io",
                    },
                ]

        :param short_name: A name for the widget, no longer than 30 characters.
        :param styles: A dictionary with keys ``"backgroundColor"`` and
            ``"headerColor"``, and values of hex colors. For example,
            ``{"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}``.

        :returns: The created :class:`.ImageWidget`.

        .. _reddit docs: https://www.reddit.com/dev/api#POST_api_widget

        Example usage:

        .. code-block:: python

            subreddit = await reddit.subreddit("test")
            widget_moderation = subreddit.widgets.mod
            image_paths = ["/path/to/image1.jpg", "/path/to/image2.png"]
            image_data = [
                {
                    "width": 600,
                    "height": 450,
                    "linkUrl": "",
                    "url": widget_moderation.upload_image(img_path),
                }
                for img_path in image_paths
            ]
            styles = {"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}
            new_widget = await widget_moderation.add_image_widget(
                short_name="My cool pictures", data=image_data, styles=styles
            )

        """
        image_widget = {
            "data": data,
            "kind": "image",
            "shortName": short_name,
            "styles": styles,
        }
        image_widget.update(other_settings)
        return await self._create_widget(image_widget)

    @_deprecate_args("data")
    async def add_menu(
        self,
        *,
        data: List[Dict[str, Union[List[Dict[str, str]], str]]],
        **other_settings,
    ) -> "asyncpraw.models.Menu":
        """Add and return a :class:`.Menu` widget.

        :param data: A list of dictionaries describing menu contents, as specified in
            `Reddit docs`_. As of this writing, the format is:

            .. code-block:: text

                [
                    {
                        "text": a string no longer than 20 characters,
                        "url": a valid URL
                    },

                    OR

                    {
                        "children": [
                            {
                                "text": a string no longer than 20 characters,
                                "url": a valid URL,
                            },
                            ...
                        ],
                        "text": a string no longer than 20 characters,
                    },
                    ...
                ]


        :returns: The created :class:`.Menu`.

        .. _reddit docs: https://www.reddit.com/dev/api#POST_api_widget

        Example usage:

        .. code-block:: python

            subreddit = await reddit.subreddit("test")
            widget_moderation = subreddit.widgets.mod
            menu_contents = [
                {"text": "My homepage", "url": "https://example.com"},
                {
                    "text": "Python packages",
                    "children": [
                        {"text": "asyncpraw", "url": "https://asyncpraw.readthedocs.io/"},
                        {"text": "requests", "url": "https://docs.python-requests.org/"},
                    ],
                },
                {"text": "Reddit homepage", "url": "https://reddit.com"},
            ]
            new_widget = await widget_moderation.add_menu(data=menu_contents)

        """
        menu = {"data": data, "kind": "menu"}
        menu.update(other_settings)
        return await self._create_widget(menu)

    @_deprecate_args("short_name", "display", "order", "styles")
    async def add_post_flair_widget(
        self,
        *,
        display: str,
        order: List[str],
        short_name: str,
        styles: Dict[str, str],
        **other_settings,
    ) -> "asyncpraw.models.PostFlairWidget":
        """Add and return a :class:`.PostFlairWidget`.

        :param display: Display style. Either ``"cloud"`` or ``"list"``.
        :param order: A list of flair template IDs. You can get all flair template IDs
            in a subreddit with:

            .. code-block:: python

                flairs = [f["id"] for f in subreddit.flair.link_templates]

        :param short_name: A name for the widget, no longer than 30 characters.
        :param styles: A dictionary with keys ``"backgroundColor"`` and
            ``"headerColor"``, and values of hex colors. For example,
            ``{"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}``.

        :returns: The created :class:`.PostFlairWidget`.

        Example usage:

        .. code-block:: python

            subreddit = await reddit.subreddit("test")
            widget_moderation = subreddit.widgets.mod
            flairs = [f["id"] async for f in subreddit.flair.link_templates]
            styles = {"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}
            new_widget = await widget_moderation.add_post_flair_widget(
                short_name="Some flairs", display="list", order=flairs, styles=styles
            )

        """
        post_flair = {
            "kind": "post-flair",
            "display": display,
            "shortName": short_name,
            "order": order,
            "styles": styles,
        }
        post_flair.update(other_settings)
        return await self._create_widget(post_flair)

    @_deprecate_args("short_name", "text", "styles")
    async def add_text_area(
        self, *, short_name: str, styles: Dict[str, str], text: str, **other_settings
    ) -> "asyncpraw.models.TextArea":
        """Add and return a :class:`.TextArea` widget.

        :param short_name: A name for the widget, no longer than 30 characters.
        :param styles: A dictionary with keys ``"backgroundColor"`` and
            ``"headerColor"``, and values of hex colors. For example,
            ``{"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}``.
        :param text: The Markdown text displayed in the widget.

        :returns: The created :class:`.TextArea`.

        Example usage:

        .. code-block:: python

            subreddit = await reddit.subreddit("test")
            widget_moderation = subreddit.widgets.mod
            styles = {"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}
            new_widget = await widget_moderation.add_text_area(
                short_name="My cool title", text="*Hello* **world**!", styles=styles
            )

        """
        text_area = {
            "shortName": short_name,
            "text": text,
            "styles": styles,
            "kind": "textarea",
        }
        text_area.update(other_settings)
        return await self._create_widget(text_area)

    @_deprecate_args("new_order", "section")
    async def reorder(
        self, new_order: List[Union[WidgetType, str]], *, section: str = "sidebar"
    ):
        """Reorder the widgets.

        :param new_order: A list of widgets. Represented as a list that contains
            :class:`.Widget` objects, or widget IDs as strings. These types may be
            mixed.
        :param section: The section to reorder (default: ``"sidebar"``).

        Example usage:

        .. code-block:: python

            subreddit = await reddit.subreddit("test")
            widgets = [widget async for widget in subreddit.widgets]
            order = list(widgets.sidebar)
            order.reverse()
            await widgets.mod.reorder(order)

        """
        order = [
            thing.id if isinstance(thing, Widget) else str(thing) for thing in new_order
        ]
        path = API_PATH["widget_order"].format(
            subreddit=self._subreddit, section=section
        )
        await self._reddit.patch(path, data={"json": dumps(order), "section": section})

    async def upload_image(self, file_path: str) -> str:
        """Upload an image to Reddit and get the URL.

        :param file_path: The path to the local file.

        :returns: The URL of the uploaded image as a ``str``.

        This method is used to upload images for widgets. For example, it can be used in
        conjunction with :meth:`.add_image_widget`, :meth:`.add_custom_widget`, and
        :meth:`.add_button_widget`.

        Example usage:

        .. code-block:: python

            my_sub = await reddit.subreddit("test")
            image_url = await my_sub.widgets.mod.upload_image("/path/to/image.jpg")
            image_data = [{"width": 300, "height": 300, "url": image_url, "linkUrl": ""}]
            styles = {"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}
            await my_sub.widgets.mod.add_image_widget(
                short_name="My cool pictures", data=image_data, styles=styles
            )

        """
        img_data = {
            "filepath": os.path.basename(file_path),
            "mimetype": "image/jpeg",
        }
        if file_path.lower().endswith(".png"):
            img_data["mimetype"] = "image/png"

        url = API_PATH["widget_lease"].format(subreddit=self._subreddit)
        # until we learn otherwise, assume this request always succeeds
        response = await self._reddit.post(url, data=img_data)
        upload_lease = response["s3UploadLease"]
        upload_data = {item["name"]: item["value"] for item in upload_lease["fields"]}
        upload_url = f"https:{upload_lease['action']}"

        with open(file_path, "rb") as image:
            upload_data["file"] = image
            response = await self._reddit._core._requestor._http.post(
                upload_url, data=upload_data
            )
        response.raise_for_status()

        return f"{upload_url}/{upload_data['key']}"


class Widget(AsyncPRAWBase):
    """Base class to represent a :class:`.Widget`."""

    @cachedproperty
    def mod(self) -> "asyncpraw.models.WidgetModeration":
        """Get an instance of :class:`.WidgetModeration` for this widget.

        .. note::

            Using any of the methods of :class:`.WidgetModeration` will likely make
            outdated the data in the :class:`.SubredditWidgets` that this widget belongs
            to. To remedy this, call :meth:`~.SubredditWidgets.refresh`.

        """
        return WidgetModeration(self, self.subreddit, self._reddit)

    def __eq__(self, other: Any) -> bool:
        """Check equality against another object."""
        if isinstance(other, Widget):
            return self.id.lower() == other.id.lower()
        return str(other).lower() == self.id.lower()

    # pylint: disable=invalid-name
    def __init__(self, reddit: "asyncpraw.Reddit", _data: Dict[str, Any]):
        """Initialize a :class:`.Widget` instance."""
        self.subreddit = ""  # in case it isn't in _data
        self.id = ""  # in case it isn't in _data
        super().__init__(reddit, _data=_data)
        self._mod = None


class ButtonWidget(Widget, BaseList):
    r"""Class to represent a widget containing one or more buttons.

    Find an existing one:

    .. code-block:: python

        button_widget = None
        subreddit = await reddit.subreddit("test")
        widgets = subreddit.widgets
        async for widget in widgets.sidebar():
            if isinstance(widget, asyncpraw.models.ButtonWidget):
                button_widget = widget
                break

        for button in button_widget:
            print(button.text, button.url)

    Create one:

    .. code-block:: python

        subreddit = await reddit.subreddit("test")
        widgets = subreddit.widgets
        buttons = [
            {
                "kind": "text",
                "text": "View source",
                "url": "https://github.com/asyncpraw-dev/asyncpraw",
                "color": "#FF0000",
                "textColor": "#00FF00",
                "fillColor": "#0000FF",
                "hoverState": {
                    "kind": "text",
                    "text": "ecruos weiV",
                    "color": "#000000",
                    "textColor": "#FFFFFF",
                    "fillColor": "#0000FF",
                },
            },
            {
                "kind": "text",
                "text": "View documentation",
                "url": "https://asyncpraw.readthedocs.io",
                "color": "#FFFFFF",
                "textColor": "#FFFF00",
                "fillColor": "#0000FF",
            },
        ]
        styles = {"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}
        button_widget = await widgets.mod.add_button_widget(
            "Things to click", "Click some of these *cool* links!", buttons, styles
        )

    For more information on creation, see :meth:`.add_button_widget`.

    Update one:

    .. code-block:: python

        new_styles = {"backgroundColor": "#FFFFFF", "headerColor": "#FF9900"}
        button_widget = await button_widget.mod.update(
            shortName="My fav buttons", styles=new_styles
        )

    Delete one:

    .. code-block:: python

        await button_widget.mod.delete()

    .. include:: ../../typical_attributes.rst

    ==================== ==============================================================
    Attribute            Description
    ==================== ==============================================================
    ``buttons``          A list of :class:`.Button`\ s. These can also be accessed just
                         by iterating over the :class:`.ButtonWidget` (e.g., ``for
                         button in button_widget``).
    ``description``      The description, in Markdown.
    ``description_html`` The description, in HTML.
    ``id``               The widget ID.
    ``kind``             The widget kind (always ``"button"``).
    ``shortName``        The short name of the widget.
    ``styles``           A ``dict`` with the keys ``"backgroundColor"`` and
                         ``"headerColor"``.
    ``subreddit``        The :class:`.Subreddit` the button widget belongs to.
    ==================== ==============================================================

    """

    CHILD_ATTRIBUTE = "buttons"


class Calendar(Widget):
    """Class to represent a calendar widget.

    Find an existing one:

    .. code-block:: python

        calendar = None
        subreddit = await reddit.subreddit("test")
        widgets = subreddit.widgets
        async for widget in widgets.sidebar():
            if isinstance(widget, asyncpraw.models.Calendar):
                calendar = widget
                break

        print(calendar.googleCalendarId)

    Create one:

    .. code-block:: python

        subreddit = await reddit.subreddit("test")
        widgets = subreddit.widgets
        styles = {"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}
        config = {
            "numEvents": 10,
            "showDate": True,
            "showDescription": False,
            "showLocation": False,
            "showTime": True,
            "showTitle": True,
        }
        cal_id = "y6nm89jy427drk8l71w75w9wjn@group.calendar.google.com"
        calendar = await widgets.mod.add_calendar(
            short_name="Upcoming Events",
            google_calendar_id=cal_id,
            requires_sync=True,
            configuration=config,
            styles=styles,
        )

    For more information on creation, see :meth:`.add_calendar`.

    Update one:

    .. code-block:: python

        new_styles = {"backgroundColor": "#FFFFFF", "headerColor": "#FF9900"}
        calendar = await calendar.mod.update(shortName="My fav events", styles=new_styles)

    Delete one:

    .. code-block:: python

        await calendar.mod.delete()

    .. include:: ../../typical_attributes.rst

    ================= =====================================================
    Attribute         Description
    ================= =====================================================
    ``configuration`` A ``dict`` describing the calendar configuration.
    ``data``          A list of dictionaries that represent events.
    ``id``            The widget ID.
    ``kind``          The widget kind (always ``"calendar"``).
    ``requiresSync``  A ``bool`` representing whether the calendar requires
                      synchronization.
    ``shortName``     The short name of the widget.
    ``styles``        A ``dict`` with the keys ``"backgroundColor"`` and
                      ``"headerColor"``.
    ``subreddit``     The :class:`.Subreddit` the button widget belongs to.
    ================= =====================================================

    """


class CommunityList(Widget, BaseList):
    r"""Class to represent a Related Communities widget.

    Find an existing one:

    .. code-block:: python

        community_list = None
        subreddit = await reddit.subreddit("test")
        widgets = subreddit.widgets
        async for widget in widgets.sidebar():
            if isinstance(widget, asyncpraw.models.CommunityList):
                community_list = widget
                break

        print(community_list)

    Create one:

    .. code-block:: python

        subreddit = await reddit.subreddit("test")
        widgets = subreddit.widgets
        styles = {"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}
        new_subreddit = await reddit.subreddit("test")
        subreddits = ["learnpython", new_subreddit]
        community_list = await widgets.mod.add_community_list(
            short_name="Related subreddits",
            data=subreddits,
            styles=styles,
            description="description",
        )

    For more information on creation, see :meth:`.add_community_list`.

    Update one:

    .. code-block:: python

        new_styles = {"backgroundColor": "#FFFFFF", "headerColor": "#FF9900"}
        community_list = await community_list.mod.update(
            shortName="My fav subs", styles=new_styles
        )

    Delete one:

    .. code-block:: python

        await community_list.mod.delete()

    .. include:: ../../typical_attributes.rst

    ============= =====================================================================
    Attribute     Description
    ============= =====================================================================
    ``data``      A list of :class:`.Subreddit`\ s. These can also be iterated over by
                  iterating over the :class:`.CommunityList` (e.g., ``for sub in
                  community_list``).
    ``id``        The widget ID.
    ``kind``      The widget kind (always ``"community-list"``).
    ``shortName`` The short name of the widget.
    ``styles``    A ``dict`` with the keys ``"backgroundColor"`` and ``"headerColor"``.
    ``subreddit`` The :class:`.Subreddit` the button widget belongs to.
    ============= =====================================================================

    """

    CHILD_ATTRIBUTE = "data"


class CustomWidget(Widget):
    """Class to represent a custom widget.

    Find an existing one:

    .. code-block:: python

        custom = None
        subreddit = await reddit.subreddit("test")
        widgets = subreddit.widgets
        async for widget in widgets.sidebar():
            if isinstance(widget, asyncpraw.models.CustomWidget):
                custom = widget
                break

        print(custom.text)
        print(custom.css)

    Create one:

    .. code-block:: python

        subreddit = await reddit.subreddit("test")
        widgets = subreddit.widgets
        styles = {"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}
        custom = await widgets.mod.add_custom_widget(
            short_name="My custom widget",
            text="# Hello world!",
            css="/**/",
            height=200,
            image_data=[],
            styles=styles,
        )

    For more information on creation, see :meth:`.add_custom_widget`.

    Update one:

    .. code-block:: python

        new_styles = {"backgroundColor": "#FFFFFF", "headerColor": "#FF9900"}
        custom = await custom.mod.update(shortName="My fav customization", styles=new_styles)

    Delete one:

    .. code-block:: python

        await custom.mod.delete()

    .. include:: ../../typical_attributes.rst

    ================= ============================================================
    Attribute         Description
    ================= ============================================================
    ``css``           The CSS of the widget, as a ``str``.
    ``height``        The height of the widget, as an ``int``.
    ``id``            The widget ID.
    ``imageData``     A ``list`` of :class:`.ImageData` that belong to the widget.
    ``kind``          The widget kind (always ``"custom"``).
    ``shortName``     The short name of the widget.
    ``styles``        A ``dict`` with the keys ``"backgroundColor"`` and
                      ``"headerColor"``.
    ``stylesheetUrl`` A link to the widget's stylesheet.
    ``subreddit``     The :class:`.Subreddit` the button widget belongs to.
    ``text``          The text contents, as Markdown.
    ``textHtml``      The text contents, as HTML.
    ================= ============================================================

    """

    def __init__(self, reddit: "asyncpraw.Reddit", _data: Dict[str, Any]):
        """Initialize a :class:`.CustomWidget` instance."""
        _data["imageData"] = [
            ImageData(reddit, data) for data in _data.pop("imageData")
        ]
        super().__init__(reddit, _data=_data)


class IDCard(Widget):
    """Class to represent an ID card widget.

    .. code-block:: python

        subreddit = await reddit.subreddit("test")
        widgets = subreddit.widgets
        id_card = await widgets.id_card()
        print(id_card.subscribersText)

    Update one:

    .. code-block:: python

        id_card = await widgets.id_card()
        await id_card.mod.update(currentlyViewingText="Bots")

    .. include:: ../../typical_attributes.rst

    ========================= =======================================================
    Attribute                 Description
    ========================= =======================================================
    ``currentlyViewingCount`` The number of redditors viewing the subreddit.
    ``currentlyViewingText``  The text displayed next to the view count. For example,
                              ``"users online"``.
    ``description``           The subreddit description.
    ``id``                    The widget ID.
    ``kind``                  The widget kind (always ``"id-card"``).
    ``shortName``             The short name of the widget.
    ``styles``                A ``dict`` with the keys ``"backgroundColor"`` and
                              ``"headerColor"``.
    ``subreddit``             The :class:`.Subreddit` the button widget belongs to.
    ``subscribersCount``      The number of subscribers to the subreddit.
    ``subscribersText``       The text displayed next to the subscriber count. For
                              example, "users subscribed".
    ========================= =======================================================

    """


class ImageWidget(Widget, BaseList):
    r"""Class to represent an image widget.

    Find an existing one:

    .. code-block:: python

        image_widget = None
        subreddit = await reddit.subreddit("test")
        widgets = subreddit.widgets
        async for widget in widgets.sidebar():
            if isinstance(widget, asyncpraw.models.ImageWidget):
                image_widget = widget
                break

        for image in image_widget:
            print(image.url)

    Create one:

    .. code-block:: python

        subreddit = await reddit.subreddit("test")
        widgets = subreddit.widgets
        image_paths = ["/path/to/image1.jpg", "/path/to/image2.png"]
        image_data = [
            {
                "width": 600,
                "height": 450,
                "linkUrl": "",
                "url": await widgets.mod.upload_image(img_path),
            }
            for img_path in image_paths
        ]
        styles = {"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}
        image_widget = await widgets.mod.add_image_widget(
            short_name="My cool pictures", data=image_data, styles=styles
        )

    For more information on creation, see :meth:`.add_image_widget`.

    Update one:

    .. code-block:: python

        new_styles = {"backgroundColor": "#FFFFFF", "headerColor": "#FF9900"}
        image_widget = await image_widget.mod.update(
            shortName="My fav images", styles=new_styles
        )

    Delete one:

    .. code-block:: python

        await image_widget.mod.delete()

    .. include:: ../../typical_attributes.rst

    ============= =====================================================================
    Attribute     Description
    ============= =====================================================================
    ``data``      A list of the :class:`.Image`\ s in this widget. Can be iterated over
                  by iterating over the :class:`.ImageWidget` (e.g., ``for img in
                  image_widget``).
    ``id``        The widget ID.
    ``kind``      The widget kind (always ``"image"``).
    ``shortName`` The short name of the widget.
    ``styles``    A ``dict`` with the keys ``"backgroundColor"`` and ``"headerColor"``.
    ``subreddit`` The :class:`.Subreddit` the button widget belongs to.
    ============= =====================================================================

    """

    CHILD_ATTRIBUTE = "data"


class Menu(Widget, BaseList):
    r"""Class to represent the top menu widget of a :class:`.Subreddit`.

    Menus can generally be found as the first item in a :class:`.Subreddit`'s top bar.

    .. code-block:: python

        subreddit = await reddit.subreddit("test")
        topbar = [widget async for widget in subreddit.widgets.topbar()]
        if len(topbar) > 0:
            probably_menu = topbar[0]
            assert isinstance(probably_menu, asyncpraw.models.Menu)
            for item in probably_menu:
                if isinstance(item, asyncpraw.models.Submenu):
                    print(item.text)
                    for child in item:
                        print("\t", child.text, child.url)
                else:  # MenuLink
                    print(item.text, item.url)

    Create one:

    .. code-block:: python

        subreddit = await reddit.subreddit("test")
        widgets = subreddit.widgets
        menu_contents = [
            {"text": "My homepage", "url": "https://example.com"},
            {
                "text": "Python packages",
                "children": [
                    {"text": "asyncpraw", "url": "https://asyncpraw.readthedocs.io/"},
                    {"text": "requests", "url": "http://python-requests.org"},
                ],
            },
            {"text": "Reddit homepage", "url": "https://reddit.com"},
        ]
        menu = await widgets.mod.add_menu(data=menu_contents)

    For more information on creation, see :meth:`.add_menu`.

    Update one:

    .. code-block:: python

        menu_items = list(menu)
        menu_items.reverse()
        menu = await menu.mod.update(data=menu_items)

    Delete one:

    .. code-block:: python

        await menu.mod.delete()

    .. include:: ../../typical_attributes.rst

    ============= ====================================================================
    Attribute     Description
    ============= ====================================================================
    ``data``      A list of the :class:`.MenuLink`\ s and :class:`.Submenu`\ s in this
                  widget. Can be iterated over by iterating over the :class:`.Menu`
                  (e.g., ``for item in menu``).
    ``id``        The widget ID.
    ``kind``      The widget kind (always ``"menu"``).
    ``subreddit`` The :class:`.Subreddit` the button widget belongs to.
    ============= ====================================================================

    """

    CHILD_ATTRIBUTE = "data"


class ModeratorsWidget(Widget, BaseList):
    r"""Class to represent a moderators widget.

    .. code-block:: python

        subreddit = await reddit.subreddit("test")
        widgets = subreddit.widgets
        print(await widgets.moderators_widget())

    Update one:

    .. code-block:: python

        new_styles = {"backgroundColor": "#FFFFFF", "headerColor": "#FF9900"}
        moderator_widget = await widgets.moderators_widget()
        await moderator_widget.mod.update(styles=new_styles)

    .. include:: ../../typical_attributes.rst

    ============= =====================================================================
    Attribute     Description
    ============= =====================================================================
    ``id``        The widget ID.
    ``kind``      The widget kind (always ``"moderators"``).
    ``mods``      A list of the :class:`.Redditor`\ s that moderate the subreddit. Can
                  be iterated over by iterating over the :class:`.ModeratorsWidget`
                  (e.g., ``for mod in widgets.moderators_widget``).
    ``styles``    A ``dict`` with the keys ``"backgroundColor"`` and ``"headerColor"``.
    ``subreddit`` The :class:`.Subreddit` the button widget belongs to.
    ``totalMods`` The total number of moderators in the subreddit.
    ============= =====================================================================

    """

    CHILD_ATTRIBUTE = "mods"

    def __init__(self, reddit: "asyncpraw.Reddit", _data: Dict[str, Any]):
        """Initialize a :class:`.ModeratorsWidget` instance."""
        if self.CHILD_ATTRIBUTE not in _data:
            # .mod.update() sometimes returns payload without "mods" field
            _data[self.CHILD_ATTRIBUTE] = []
        super().__init__(reddit, _data=_data)


class PostFlairWidget(Widget, BaseList):
    """Class to represent a post flair widget.

    Find an existing one:

    .. code-block:: python

        post_flair_widget = None
        subreddit = await reddit.subreddit("test")
        widgets = subreddit.widgets
        async for widget in widgets.sidebar():
            if isinstance(widget, asyncpraw.models.PostFlairWidget):
                post_flair_widget = widget
                break

        async for flair in post_flair_widget:
            print(flair)
            print(post_flair_widget.templates[flair])

    Create one:

    .. code-block:: python

        subreddit = await reddit.subreddit("test")
        widgets = subreddit.widgets
        flairs = [f["id"] async for f in subreddit.flair.link_templates]
        styles = {"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}
        post_flair = await widgets.mod.add_post_flair_widget(
            short_name="Some flairs", display="list", order=flairs, styles=styles
        )

    For more information on creation, see :meth:`.add_post_flair_widget`.

    Update one:

    .. code-block:: python

        new_styles = {"backgroundColor": "#FFFFFF", "headerColor": "#FF9900"}
        post_flair = await post_flair.mod.update(shortName="My fav flairs", styles=new_styles)

    Delete one:

    .. code-block:: python

        await post_flair.mod.delete()

    .. include:: ../../typical_attributes.rst

    ============= =====================================================================
    Attribute     Description
    ============= =====================================================================
    ``display``   The display style of the widget, either ``"cloud"`` or ``"list"``.
    ``id``        The widget ID.
    ``kind``      The widget kind (always ``"post-flair"``).
    ``order``     A list of the flair IDs in this widget. Can be iterated over by
                  iterating over the :class:`.PostFlairWidget` (e.g., ``for flair_id in
                  post_flair``).
    ``shortName`` The short name of the widget.
    ``styles``    A ``dict`` with the keys ``"backgroundColor"`` and ``"headerColor"``.
    ``subreddit`` The :class:`.Subreddit` the button widget belongs to.
    ``templates`` A ``dict`` that maps flair IDs to dictionaries that describe flairs.
    ============= =====================================================================

    """

    CHILD_ATTRIBUTE = "order"


class RulesWidget(Widget, BaseList):
    """Class to represent a rules widget.

    .. code-block:: python

        subreddit = await reddit.subreddit("test")
        widgets = subreddit.widgets
        rules_widget = None
        async for widget in widgets.sidebar():
            if isinstance(widget, asyncpraw.models.RulesWidget):
                rules_widget = widget
                break
        from pprint import pprint

        pprint(rules_widget.data)

    Update one:

    .. code-block:: python

        new_styles = {"backgroundColor": "#FFFFFF", "headerColor": "#FF9900"}
        await rules_widget.mod.update(
            display="compact", shortName="The LAWS", styles=new_styles
        )

    .. include:: ../../typical_attributes.rst

    ============= =====================================================================
    Attribute     Description
    ============= =====================================================================
    ``data``      A list of the subreddit rules. Can be iterated over by iterating over
                  the :class:`.RulesWidget` (e.g., ``for rule in rules_widget``).
    ``display``   The display style of the widget, either ``"full"`` or ``"compact"``.
    ``id``        The widget ID.
    ``kind``      The widget kind (always ``"subreddit-rules"``).
    ``shortName`` The short name of the widget.
    ``styles``    A ``dict`` with the keys ``"backgroundColor"`` and ``"headerColor"``.
    ``subreddit`` The :class:`.Subreddit` the button widget belongs to.
    ============= =====================================================================

    """

    CHILD_ATTRIBUTE = "data"

    def __init__(self, reddit: "asyncpraw.Reddit", _data: Dict[str, Any]):
        """Initialize a :class:`.RulesWidget` instance."""
        if self.CHILD_ATTRIBUTE not in _data:
            # .mod.update() sometimes returns payload without "data" field
            _data[self.CHILD_ATTRIBUTE] = []
        super().__init__(reddit, _data=_data)


class TextArea(Widget):
    """Class to represent a text area widget.

    Find a text area in a subreddit:

    .. code-block:: python

        subreddit = await reddit.subreddit("test")
        widgets = subreddit.widgets
        text_area = None
        async for widget in widgets.sidebar():
            if isinstance(widget, asyncpraw.models.TextArea):
                text_area = widget
                break
        print(text_area.text)

    Create one:

    .. code-block:: python

        subreddit = await reddit.subreddit("test")
        widgets = subreddit.widgets
        styles = {"backgroundColor": "#FFFF66", "headerColor": "#3333EE"}
        text_area = await widgets.mod.add_text_area(
            short_name="My cool title", text="*Hello* **world**!", styles=styles
        )

    For more information on creation, see :meth:`.add_text_area`.

    Update one:

    .. code-block:: python

        new_styles = {"backgroundColor": "#FFFFFF", "headerColor": "#FF9900"}
        text_area = await text_area.mod.update(shortName="My fav text", styles=new_styles)

    Delete one:

    .. code-block:: python

        await text_area.mod.delete()

    .. include:: ../../typical_attributes.rst

    ============= =====================================================================
    Attribute     Description
    ============= =====================================================================
    ``id``        The widget ID.
    ``kind``      The widget kind (always ``"textarea"``).
    ``shortName`` The short name of the widget.
    ``styles``    A ``dict`` with the keys ``"backgroundColor"`` and ``"headerColor"``.
    ``subreddit`` The :class:`.Subreddit` the button widget belongs to.
    ``text``      The widget's text, as Markdown.
    ``textHtml``  The widget's text, as HTML.
    ============= =====================================================================

    """


class WidgetEncoder(JSONEncoder):
    """Class to encode widget-related objects."""

    def default(self, o: Any) -> Any:  # pylint: disable=E0202
        """Serialize ``AsyncPRAWBase`` objects."""
        if isinstance(o, self._subreddit_class):
            return str(o)
        elif isinstance(o, AsyncPRAWBase):
            return {key: val for key, val in vars(o).items() if not key.startswith("_")}
        return JSONEncoder.default(self, o)


class WidgetModeration:
    """Class for moderating a particular widget.

    Example usage:

    .. code-block:: python

        subreddit = await reddit.subreddit("test")
        sidebar = [widget async for widget in subreddit.widgets.sidebar()]
        widget = sidebar[0]
        await widget.mod.update(shortName="My new title")
        await widget.mod.delete()

    """

    def __init__(
        self,
        widget: "asyncpraw.models.Widget",
        subreddit: Union["asyncpraw.models.Subreddit", str],
        reddit: "asyncpraw.Reddit",
    ):
        """Initialize a :class:`.WidgetModeration` instance."""
        self.widget = widget
        self._reddit = reddit
        self._subreddit = subreddit

    async def delete(self):
        """Delete the widget.

        Example usage:

        .. code-block:: python

            await widget.mod.delete()

        """
        path = API_PATH["widget_modify"].format(
            widget_id=self.widget.id, subreddit=self._subreddit
        )
        await self._reddit.delete(path)

    async def update(self, **kwargs) -> WidgetType:
        """Update the widget. Returns the updated widget.

        Parameters differ based on the type of widget. See `Reddit documentation
        <https://www.reddit.com/dev/api#PUT_api_widget_{widget_id}>`_ or the document of
        the particular type of widget.

        :returns: The updated :class:`.Widget`.

        For example, update a text widget like so:

        .. code-block:: python

            await text_widget.mod.update(shortName="New text area", text="Hello!")

        .. note::

            Most parameters follow the ``lowerCamelCase`` convention. When in doubt,
            check the Reddit documentation linked above.

        """
        path = API_PATH["widget_modify"].format(
            widget_id=self.widget.id, subreddit=self._subreddit
        )
        payload = {
            key: value
            for key, value in vars(self.widget).items()
            if not key.startswith("_")
        }
        del payload["subreddit"]  # not JSON serializable
        if "mod" in payload:
            del payload["mod"]
        payload.update(kwargs)
        widget = await self._reddit.put(
            path, data={"json": dumps(payload, cls=WidgetEncoder)}
        )
        widget.subreddit = self._subreddit
        return widget



from asyncpraw.models import PollOption

from ... import UnitTest


class TestPollOption(UnitTest):
    def test_repr(self, reddit):
        option = PollOption(reddit, {"id": "anID", "text": "theText", "vote_count": 0})
        assert repr(option) == "PollOption(id='anID')"

    def test_str(self, reddit):
        option = PollOption(reddit, {"id": "anID", "text": "theText", "vote_count": 0})
        assert str(option) == "theText"



"""This file should be updated as files/classes/functions are deprecated."""
import inspect
from contextlib import contextmanager
from enum import IntFlag, auto

import pytest

from asyncpraw.util import _deprecate_args

from .. import UnitTest

keyword_only = {
    "args": [
        [("arg2", "arg1", "arg3", "arg0"), dict()],
        ("arg0", "arg1", "arg2", "arg3"),
    ],
    "kwargs": [
        [(), dict(arg2="arg2", arg1="arg1", arg3="arg3", arg0="arg0")],
        ("arg0", "arg1", "arg2", "arg3"),
    ],
    "mix": [
        [("arg2",), dict(arg1="arg1")],
        (None, "arg1", "arg2", None),
    ],
    "one_kwarg": [
        [(), dict(arg0="arg0")],
        ("arg0", None, None, None),
    ],
    "one_arg": [
        [("arg2",), dict()],
        (None, None, "arg2", None),
    ],
}
with_positional = {
    "args": [
        [("arg0", "arg2", "arg1", "arg3"), dict()],
        ("arg0", "arg1", "arg2", "arg3"),
    ],
    "kwargs": [
        [("arg0",), dict(arg2="arg2", arg1="arg1", arg3="arg3")],
        ("arg0", "arg1", "arg2", "arg3"),
    ],
    "mix": [
        [("arg0", "arg2"), dict(arg1="arg1", arg3=None)],
        ("arg0", "arg1", "arg2", None),
    ],
    "one_kwarg": [
        [(), dict(arg0="arg0")],
        ("arg0", None, None, None),
    ],
    "one_arg": [
        [("arg0",), dict()],
        ("arg0", None, None, None),
    ],
}


class AsyncType(IntFlag):
    ASYNC = auto()
    ASYNC_GENERATOR = auto()
    SYNC = auto()

    def __str__(self):
        return f"__{self.name.lower()}"

    def __repr__(self):
        return self.name.lower()


def _gen_warning(func, args):
    arg_list = list(map(repr, args))
    arg_count = len(args)
    plural = arg_count > 1
    arg_string = (
        " and ".join(arg_list)
        if arg_count < 3
        else f"{', '.join(arg_list[:-1])}, and {arg_list[-1]}"
    )
    arg_string += f" as {'' if plural else 'a '}"
    arg_string += f"keyword argument{'s' if plural else ''}"
    return (
        f"Positional arguments for {func.__qualname__!r} will no longer be supported in"
        f" Async PRAW 8.\nCall this function with {arg_string}."
    )


@contextmanager
def _check_warning(func, args):
    if args:
        with pytest.warns(DeprecationWarning, match=_gen_warning(func, args)):
            yield
    else:
        yield


def _prepare_args(arguments, func):
    args, kwargs = arguments
    parameters = inspect.signature(func).parameters.values()
    check_args = args[
        len(
            [
                param
                for param in parameters
                if param.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD
            ]
        ) :
    ]
    return args, check_args, kwargs


@_deprecate_args("arg2", "arg1", "arg3", "arg0")
def arg_test_global(*, arg0=None, arg1=None, arg2=None, arg3=None):
    return arg0, arg1, arg2, arg3


@_deprecate_args("arg2", "arg1", "arg3", "arg0")
async def arg_test_global__async(*, arg0=None, arg1=None, arg2=None, arg3=None):
    return arg0, arg1, arg2, arg3


@_deprecate_args("arg2", "arg1", "arg3", "arg0")
async def arg_test_global__async_generator(
    *, arg0=None, arg1=None, arg2=None, arg3=None
):
    yield arg0
    yield arg1
    yield arg2
    yield arg3


@_deprecate_args("arg0", "arg2", "arg1", "arg3")
def arg_test_global_with_positional(arg0, *, arg1=None, arg2=None, arg3=None):
    return arg0, arg1, arg2, arg3


@_deprecate_args("arg0", "arg2", "arg1", "arg3")
async def arg_test_global_with_positional__async(
    arg0, *, arg1=None, arg2=None, arg3=None
):
    return arg0, arg1, arg2, arg3


@_deprecate_args("arg0", "arg2", "arg1", "arg3")
async def arg_test_global_with_positional__async_generator(
    arg0, *, arg1=None, arg2=None, arg3=None
):
    yield arg0
    yield arg1
    yield arg2
    yield arg3


class ArgTest:
    @_deprecate_args("arg2", "arg1", "arg3", "arg0")
    def arg_test(self, *, arg0=None, arg1=None, arg2=None, arg3=None):
        return arg0, arg1, arg2, arg3

    @_deprecate_args("arg2", "arg1", "arg3", "arg0")
    async def arg_test__async(self, *, arg0=None, arg1=None, arg2=None, arg3=None):
        return arg0, arg1, arg2, arg3

    @_deprecate_args("arg2", "arg1", "arg3", "arg0")
    async def arg_test__async_generator(
        self, *, arg0=None, arg1=None, arg2=None, arg3=None
    ):
        yield arg0
        yield arg1
        yield arg2
        yield arg3

    @_deprecate_args("arg0", "arg2", "arg1", "arg3")
    def arg_test_with_positional(self, arg0, *, arg1=None, arg2=None, arg3=None):
        return arg0, arg1, arg2, arg3

    @_deprecate_args("arg0", "arg2", "arg1", "arg3")
    async def arg_test_with_positional__async(
        self, arg0, *, arg1=None, arg2=None, arg3=None
    ):
        return arg0, arg1, arg2, arg3

    @_deprecate_args("arg0", "arg2", "arg1", "arg3")
    async def arg_test_with_positional__async_generator(
        self, arg0, *, arg1=None, arg2=None, arg3=None
    ):
        yield arg0
        yield arg1
        yield arg2
        yield arg3


def pytest_generate_tests(metafunc):
    # called once per each test function
    if "positional" in metafunc.function.__name__:
        test_cases = metafunc.cls.params["positional"]
    else:
        test_cases = metafunc.cls.params["keyword"]
    name = "_".join(metafunc.function.__name__.split("_")[1:])
    cases = []
    ids = []
    for async_type in list(AsyncType):
        function_name = name
        if async_type in AsyncType.ASYNC | AsyncType.ASYNC_GENERATOR:
            function_name += str(async_type)
        for test_case, parameters in test_cases.items():
            cases.append([function_name, *parameters, async_type])
            ids.append(",".join([repr(async_type), test_case]))
    signature = inspect.signature(metafunc.function)
    args = [arg.name for arg in signature.parameters.values() if arg.name != "self"]
    metafunc.parametrize(argnames=args, argvalues=cases, ids=ids)


@pytest.mark.filterwarnings("error", category=DeprecationWarning)
class TestDeprecateArgs(UnitTest):
    def setup_method(self, _):
        self.arg_test = ArgTest()

    params = {
        "keyword": keyword_only,
        "positional": with_positional,
    }

    async def _execute_test(self, func_name, arguments, expected_results, async_type):
        if "global" in func_name:
            func = globals()[func_name]
        else:
            func = getattr(self.arg_test, func_name)
        args, check_args, kwargs = _prepare_args(arguments, func)
        with _check_warning(func, check_args):
            if async_type is AsyncType.ASYNC:
                results = await func(*args, **kwargs)
            elif async_type is AsyncType.ASYNC_GENERATOR:
                results = tuple([item async for item in func(*args, **kwargs)])
            else:
                results = func(*args, **kwargs)
        assert expected_results == results

    async def test_arg_test_global(
        self, func_name, arguments, expected_result, async_type
    ):
        await self._execute_test(func_name, arguments, expected_result, async_type)

    async def test_arg_test_global_with_positional(
        self, func_name, arguments, expected_result, async_type
    ):
        await self._execute_test(func_name, arguments, expected_result, async_type)

    async def test_arg_test(self, func_name, arguments, expected_result, async_type):
        await self._execute_test(func_name, arguments, expected_result, async_type)

    async def test_arg_test_with_positional(
        self, func_name, arguments, expected_result, async_type
    ):
        await self._execute_test(func_name, arguments, expected_result, async_type)



"""Provide the BaseList class."""
from typing import TYPE_CHECKING, Any, Dict, Iterator

from ..base import AsyncPRAWBase

if TYPE_CHECKING:  # pragma: no cover
    import asyncpraw


class BaseList(AsyncPRAWBase):
    """An abstract class to coerce a list into an :class:`.AsyncPRAWBase`."""

    CHILD_ATTRIBUTE = None

    def __init__(self, reddit: "asyncpraw.Reddit", _data: Dict[str, Any]):
        """Initialize a :class:`.BaseList` instance.

        :param reddit: An instance of :class:`.Reddit`.

        """
        super().__init__(reddit, _data=_data)

        if self.CHILD_ATTRIBUTE is None:
            raise NotImplementedError("BaseList must be extended.")

        child_list = getattr(self, self.CHILD_ATTRIBUTE)
        for index, item in enumerate(child_list):
            child_list[index] = reddit._objector.objectify(item)

    def __contains__(self, item: Any) -> bool:
        """Test if item exists in the list."""
        return item in getattr(self, self.CHILD_ATTRIBUTE)

    def __getitem__(self, index: int) -> Any:
        """Return the item at position index in the list."""
        return getattr(self, self.CHILD_ATTRIBUTE)[index]

    def __iter__(self) -> Iterator[Any]:
        """Return an iterator to the list."""
        return getattr(self, self.CHILD_ATTRIBUTE).__iter__()

    def __len__(self) -> int:
        """Return the number of items in the list."""
        return len(getattr(self, self.CHILD_ATTRIBUTE))

    def __str__(self) -> str:
        """Return a string representation of the list."""
        return str(getattr(self, self.CHILD_ATTRIBUTE))



"""A test that is run only by GitHub Actions

This test makes real network requests, so environment variables should be specified in
GitHub Actions.

"""
import os

import pytest

from asyncpraw import Reddit
from asyncpraw.models import Submission


@pytest.mark.skipif(
    not os.getenv("NETWORK_TEST_CLIENT_ID"),
    reason="Not running from the NETWORK_TEST ci task on praw-dev/asyncpraw",
)
async def test_github_actions():
    reddit = Reddit(
        client_id=os.getenv("NETWORK_TEST_CLIENT_ID"),
        client_secret=os.getenv("NETWORK_TEST_CLIENT_SECRET"),
        user_agent="GitHub Actions CI Testing",
    )
    subreddit = await reddit.subreddit("all")
    async for submission in subreddit.hot():
        assert isinstance(submission, Submission)
        break



"""Test asyncpraw.models.subreddits."""
import sys

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.models import Subreddit

from .. import IntegrationTest


class TestSubreddits(IntegrationTest):
    async def test_default(self):
        with self.use_cassette():
            subreddits = await self.async_list(
                self.reddit.subreddits.default(limit=None)
            )
        assert 0 < len(subreddits) < 100

    async def test_premium__without_premium(self):
        with self.use_cassette():
            subreddits = await self.async_list(self.reddit.subreddits.premium())
        assert len(subreddits) == 0

    async def test_premium__with_premium(self):
        with self.use_cassette():
            subreddits = await self.async_list(self.reddit.subreddits.premium())
        assert len(subreddits) == 100

    async def test_gold__without_gold(self):  # ensure backwards compatibility
        with self.use_cassette():
            subreddits = await self.async_list(self.reddit.subreddits.gold())
        assert len(subreddits) == 0

    async def test_gold__with_gold(self):  # ensure backwards compatibility
        with self.use_cassette():
            subreddits = await self.async_list(self.reddit.subreddits.gold())
        assert len(subreddits) == 100

    async def test_new(self):
        with self.use_cassette():
            subreddits = await self.async_list(self.reddit.subreddits.new(limit=300))
        assert len(subreddits) == 300

    async def test_popular(self):
        with self.use_cassette():
            subreddits = await self.async_list(self.reddit.subreddits.popular(limit=15))
        assert len(subreddits) == 15

    async def test_recommended(
        self,
    ):  # FIXME: always seems to return []; same with praw
        with self.use_cassette():
            subreddits = await self.reddit.subreddits.recommended(
                ["earthporn"], omit_subreddits=["cityporn"]
            )
        assert (
            len(subreddits) == 0
        )  # This is like this for coverage. Endpoint does not seem to work for me.
        # assert len(subreddits) > 1
        # for subreddit in subreddits:
        #     assert isinstance(subreddit, Subreddit)

    async def test_recommended__with_multiple(
        self,
    ):  # FIXME: always seems to return []; same with praw
        with self.use_cassette():
            subreddits = await self.reddit.subreddits.recommended(
                ["cityporn", "earthporn"],
                omit_subreddits=["skyporn", "winterporn"],
            )
        assert (
            len(subreddits) == 0
        )  # This is like this for coverage. Endpoint does not seem to work for me.
        # assert len(subreddits) > 1
        # for subreddit in subreddits:
        #     assert isinstance(subreddit, Subreddit)

    async def test_search(self):
        with self.use_cassette():
            found = False
            async for subreddit in self.reddit.subreddits.search("asyncpraw"):
                assert isinstance(subreddit, Subreddit)
                found = True
            assert found

    # async def test_search_by_topic(self): # FIXME: endpoint does not exist anymore
    #     with self.use_cassette():
    #         found = False
    #         async for subreddit in self.reddit.subreddits.search_by_topic("sports"):
    #             assert isinstance(subreddit, Subreddit)
    #             found = True
    #         assert found

    async def test_search_by_name(self):
        with self.use_cassette():
            subreddits = await self.async_list(
                self.reddit.subreddits.search_by_name("reddit")
            )
            assert isinstance(subreddits, list)
            assert len(subreddits) > 1
            assert all(isinstance(x, Subreddit) for x in subreddits)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_stream(self, _):
        with self.use_cassette():
            generator = self.reddit.subreddits.stream()
            for i in range(101):
                assert isinstance(await self.async_next(generator), Subreddit)



from unittest import mock

import pytest

from .. import IntegrationTest


class TestModNotes(IntegrationTest):
    REDDITOR = "pyapitestuser3"

    async def test_create_note__submission(self):
        self.reddit.read_only = False
        with self.use_cassette():
            submission = await self.reddit.submission("uflrmv")
            result_note = await submission.mod.create_note(
                label="HELPFUL_USER", note="test note"
            )
            assert result_note.user == submission.author
            assert result_note.note == "test note"
            assert result_note.label == "HELPFUL_USER"

    async def test_create_note__thing_fullname(self):
        self.reddit.read_only = False
        with self.use_cassette():
            submission = await self.reddit.submission("uflrmv")
            result_note = await self.reddit.notes.create(
                label="HELPFUL_USER", note="test note", thing=submission.fullname
            )
            assert result_note.user == submission.author
            assert result_note.id.startswith("ModNote")
            assert result_note.moderator.name == pytest.placeholders.username
            assert result_note.note == "test note"
            assert result_note.label == "HELPFUL_USER"
            assert result_note.reddit_id == submission.fullname

    async def test_create_note__thing_submission(self):
        self.reddit.read_only = False
        with self.use_cassette("TestModNotes.test_create_note__submission"):
            submission = await self.reddit.submission("uflrmv")
            result_note = await self.reddit.notes.create(
                label="HELPFUL_USER", note="test note", thing=submission
            )
            assert result_note.user == submission.author
            assert result_note.note == "test note"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_delete_note(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            result_note = await subreddit.mod.notes.create(
                redditor=self.REDDITOR, note="test note"
            )
            await subreddit.mod.notes.delete(
                note_id=result_note.id, redditor=result_note.user
            )
            notes = await self.async_list(subreddit.mod.notes.redditors(self.REDDITOR))
            assert result_note not in notes

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_delete_note__all_notes(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            result_note = await subreddit.mod.notes.create(
                redditor=self.REDDITOR, note="test note"
            )
            await subreddit.mod.notes.delete(delete_all=True, redditor=result_note.user)
            notes = await self.async_list(subreddit.mod.notes.redditors(self.REDDITOR))
            assert len(notes) == 0



"""asyncpraw setup.py"""
import os
import re
from codecs import open
from os import path

from setuptools import find_packages, setup

PACKAGE_NAME = "asyncpraw"
HERE = path.abspath(path.dirname(__file__))
with open(path.join(HERE, "README.rst"), encoding="utf-8") as fp:
    README = fp.read()
with open(path.join(HERE, PACKAGE_NAME, "const.py"), encoding="utf-8") as fp:
    VERSION = re.search('__version__ = "([^"]+)"', fp.read()).group(1)

extras = {
    "ci": ["coveralls"],
    "dev": ["packaging"],
    "lint": ["pre-commit"],
    "readthedocs": [
        "sphinx",
        "sphinx-rtd-dark-mode",
        "sphinx_rtd_theme",
        "sphinxcontrib-trio",
    ],
    "test": [
        "asynctest >=0.13.0 ; python_version < '3.8'",
        "mock >=0.8",
        "pytest ==7.2.*",
        "pytest-asyncio",
        "pytest-vcr",
        "testfixtures >4.13.2, <7",
        "vcrpy >=4.1.1"
        if os.getenv("PYPI_UPLOAD", False)
        else "vcrpy@git+https://github.com/kevin1024/vcrpy.git@b1bc5c3a02db0447c28ab9a4cee314aeb6cdf1a7",
    ],
}
extras["lint"] += extras["readthedocs"]
extras["dev"] += extras["lint"] + extras["test"]

setup(
    name=PACKAGE_NAME,
    author="Joel Payne",
    author_email="lilspazjoekp@gmail.com",
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Utilities",
    ],
    description=(
        'Async PRAW, an abbreviation for "Asynchronous Python Reddit API Wrapper", is a'
        " python package that allows for simple access to Reddit's API."
    ),
    extras_require=extras,
    install_requires=[
        "aiofiles <1",
        "aiohttp <4",
        "aiosqlite <=0.17.0",
        "asyncio_extras <=1.3.2",
        "asyncprawcore >=2.1, <3",
        "update_checker >=0.18",
    ],
    keywords="reddit api wrapper asyncpraw praw async asynchronous",
    license="Simplified BSD License",
    long_description=README,
    package_data={
        "": ["LICENSE.txt", "praw_license.txt"],
        PACKAGE_NAME: ["*.ini", "images/*.png"],
    },
    packages=find_packages(exclude=["tests", "tests.*", "tools", "tools.*"]),
    project_urls={
        "Change Log": "https://asyncpraw.readthedocs.io/en/latest/package_info/change_log.html",
        "Documentation": "https://asyncpraw.readthedocs.io/",
        "Issue Tracker": "https://github.com/praw-dev/asyncpraw/issues",
        "Source Code": "https://github.com/praw-dev/asyncpraw",
    },
    version=VERSION,
)



import pytest

from asyncpraw.models import ModmailConversation

from ... import UnitTest


class TestModmailConversation(UnitTest):
    def test_construct_failure(self, reddit):
        message = "Either 'id' or '_data' must be provided."
        with pytest.raises(TypeError) as excinfo:
            ModmailConversation(reddit)
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            ModmailConversation(reddit, "dummy", _data={"id": "dummy"})
        assert str(excinfo.value) == message

    def test_str(self, reddit):
        conversation = ModmailConversation(reddit, _data={"id": "ik72"})
        assert str(conversation) == "ik72"

        conversation = ModmailConversation(reddit, "ik72")
        assert str(conversation) == "ik72"



"""Test asyncpraw.models.LiveThread"""
import sys

import pytest

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.const import API_PATH
from asyncpraw.exceptions import RedditAPIException
from asyncpraw.models import LiveThread, LiveUpdate, Redditor, Submission

from ... import IntegrationTest


class TestLiveUpdate(IntegrationTest):
    async def test_attributes(self):
        thread = LiveThread(self.reddit, "ukaeu1ik4sw5")
        with self.use_cassette():
            update = await thread.get_update("7827987a-c998-11e4-a0b9-22000b6a88d2")
            assert isinstance(update.author, Redditor)
            assert update.author == "umbrae"
            assert update.name == "LiveUpdate_7827987a-c998-11e4-a0b9-22000b6a88d2"
            assert update.body.startswith("Small change")


class TestLiveThread(IntegrationTest):
    async def test_contributor(self):
        thread = LiveThread(self.reddit, "ukaeu1ik4sw5")
        with self.use_cassette():
            contributors = [contributor async for contributor in thread.contributor()]
        assert isinstance(contributors, list)
        assert len(contributors) > 0
        for contributor in contributors:
            assert "permissions" in contributor.__dict__
            assert isinstance(contributor, Redditor)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_contributor__with_manage_permission(self, _):
        # see issue #710 for more info
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "xyu8kmjvfrww")
        url = API_PATH["live_contributors"].format(id=thread.id)
        with self.use_cassette():
            data = await thread._reddit.request(method="GET", path=url)
            contributors = [contributor async for contributor in thread.contributor()]
        assert isinstance(data, dict)
        assert isinstance(contributors, list)
        assert len(contributors) > 0

    async def test_init(self):
        with self.use_cassette():
            thread = await self.reddit.live("ukaeu1ik4sw5", fetch=True)
            assert thread.title == "reddit updates"

    async def test_discussions(self):
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        with self.use_cassette():
            async for submission in thread.discussions(limit=None):
                assert isinstance(submission, Submission)
                assert submission.title == "reddit updates"

    async def test_report(self):
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        with self.use_cassette():
            await thread.report("spam")

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_updates(self, _):
        thread = LiveThread(self.reddit, "ukaeu1ik4sw5")
        with self.use_cassette():
            async for update in thread.updates(limit=None):
                assert update.thread == thread
        assert update.body.startswith("Small change:")


class TestLiveThreadStream(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_updates(self, _):
        with self.use_cassette():
            live_thread = await self.reddit.live("ta535s1hq2je")
            generator = live_thread.stream.updates()
            for i in range(101):
                assert isinstance(await self.async_next(generator), LiveUpdate)


class TestLiveContributorRelationship(IntegrationTest):
    async def test_accept_invite(self):
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        with self.use_cassette():
            await thread.contributor.accept_invite()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_invite__already_invited(self, _):
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        with self.use_cassette():
            with pytest.raises(RedditAPIException) as excinfo:
                await thread.contributor.invite(pytest.placeholders.username)
        assert excinfo.value.items[0].error_type == "LIVEUPDATE_ALREADY_CONTRIBUTOR"

    async def test_invite__empty_list(self):
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        with self.use_cassette():
            await thread.contributor.invite(
                pytest.placeholders.username, permissions=[]
            )

    async def test_invite__limited(self):
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        with self.use_cassette():
            await thread.contributor.invite(
                pytest.placeholders.username, permissions=["manage", "edit"]
            )

    async def test_invite__none(self):
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        with self.use_cassette():
            await thread.contributor.invite(
                pytest.placeholders.username, permissions=None
            )

    async def test_invite__redditor(self):
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        redditor = Redditor(
            self.reddit, _data={"name": pytest.placeholders.username, "id": "3ebyblla"}
        )
        with self.use_cassette():
            await thread.contributor.invite(redditor)

    async def test_leave(self):
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        with self.use_cassette():
            await thread.contributor.leave()

    async def test_remove__fullname(self):
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        with self.use_cassette():
            await thread.contributor.remove("t2_3ebyblla")

    async def test_remove__redditor(self):
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        redditor = Redditor(
            self.reddit, _data={"name": pytest.placeholders.username, "id": "3ebyblla"}
        )
        with self.use_cassette():
            await thread.contributor.remove(redditor)

    async def test_remove_invite__fullname(self):
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        with self.use_cassette():
            await thread.contributor.remove_invite("t2_3ebyblla")

    async def test_remove_invite__redditor(self):
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        redditor = Redditor(
            self.reddit, _data={"name": pytest.placeholders.username, "id": "3ebyblla"}
        )
        with self.use_cassette():
            await thread.contributor.remove_invite(redditor)

    async def test_update__empty_list(self):
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        with self.use_cassette():
            await thread.contributor.update(
                pytest.placeholders.username, permissions=[]
            )

    async def test_update__limited(self):
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        with self.use_cassette():
            await thread.contributor.update(
                pytest.placeholders.username, permissions=["manage", "edit"]
            )

    async def test_update__none(self):
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        with self.use_cassette():
            await thread.contributor.update(
                pytest.placeholders.username, permissions=None
            )

    async def test_update_invite__empty_list(self):
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        with self.use_cassette():
            await thread.contributor.update_invite(
                pytest.placeholders.username, permissions=[]
            )

    async def test_update_invite__limited(self):
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        with self.use_cassette():
            await thread.contributor.update_invite(
                pytest.placeholders.username, permissions=["manage", "edit"]
            )

    async def test_update_invite__none(self):
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        with self.use_cassette():
            await thread.contributor.update_invite(
                pytest.placeholders.username, permissions=None
            )


class TestLiveThreadContribution(IntegrationTest):
    async def test_add(self):
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        with self.use_cassette():
            await thread.contrib.add("* `LiveThreadContribution.add() test`")

    async def test_close(self):
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        with self.use_cassette():
            await thread.contrib.close()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update__partial_settings(self, _):
        old_settings = {
            "title": "old title",
            "description": "## old description",
            "nsfw": False,
            "resources": "## old resources",
        }
        new_settings = {"title": "new title", "nsfw": True}
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        with self.use_cassette():
            await thread.contrib.update(**new_settings)
            thread = await self.reddit.live("1595195m6j9zw", fetch=True)
            assert thread.title == new_settings["title"]
            assert thread.description == old_settings["description"]
            assert thread.nsfw == new_settings["nsfw"]
            assert thread.resources == old_settings["resources"]

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update__full_settings(self, _):
        new_settings = {
            "title": "new title 2",
            "description": "## new description 2",
            "nsfw": True,
            "resources": "## new resources 2",
        }
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        with self.use_cassette():
            await thread.contrib.update(**new_settings)
            thread = await self.reddit.live("1595195m6j9zw", fetch=True)
            assert thread.title == new_settings["title"]
            assert thread.description == new_settings["description"]
            assert thread.nsfw == new_settings["nsfw"]
            assert thread.resources == new_settings["resources"]

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update__other_settings(self, _):
        new_settings = {
            "title": "new title",
            "other1": "other 1",
            "other2": "other 2",
        }
        self.reddit.read_only = False
        thread = LiveThread(self.reddit, "1595195m6j9zw")
        with self.use_cassette():
            await thread.contrib.update(**new_settings)


class TestLiveUpdateContribution(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_remove(self, _):
        self.reddit.read_only = False
        update = LiveUpdate(
            self.reddit, "1595195m6j9zw", "ec5ead40-bf2a-11ea-a3be-0e0d584e0b0b"
        )
        with self.use_cassette():
            await update.contrib.remove()

    async def test_strike(self):
        self.reddit.read_only = False
        update = LiveUpdate(
            self.reddit, "1595195m6j9zw", "3e95636e-bf2c-11ea-9488-0e29bbfe5f37"
        )
        with self.use_cassette():
            await update.contrib.strike()



"""Provide the draft class."""
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

from ...const import API_PATH
from ...exceptions import ClientException
from .base import RedditBase
from .subreddit import Subreddit
from .user_subreddit import UserSubreddit

if TYPE_CHECKING:  # pragma: no cover
    import asyncpraw


class Draft(RedditBase):
    """A class that represents a Reddit submission draft.

    .. include:: ../../typical_attributes.rst

    ========================== ======================================================
    Attribute                  Description
    ========================== ======================================================
    ``link_flair_template_id`` The link flair's ID.
    ``link_flair_text``        The link flair's text content, or ``None`` if not
                               flaired.
    ``modified``               Time the submission draft was modified, represented in
                               `Unix Time`_.
    ``original_content``       Whether the submission draft will be set as original
                               content.
    ``selftext``               The submission draft's selftext. ``None`` if a link
                               submission draft.
    ``spoiler``                Whether the submission will be marked as a spoiler.
    ``subreddit``              Provides an instance of :class:`.Subreddit` or
                               :class:`.UserSubreddit` (if set).
    ``title``                  The title of the submission draft.
    ``url``                    The URL the submission draft links to.
    ========================== ======================================================

    .. _unix time: https://en.wikipedia.org/wiki/Unix_time

    """

    STR_FIELD = "id"

    @classmethod
    async def _prepare_data(
        cls,
        *,
        flair_id: Optional[str] = None,
        flair_text: Optional[str] = None,
        is_public_link: Optional[bool] = None,
        nsfw: Optional[bool] = None,
        original_content: Optional[bool] = None,
        selftext: Optional[str] = None,
        send_replies: Optional[bool] = None,
        spoiler: Optional[bool] = None,
        subreddit: Optional[
            Union["asyncpraw.models.Subreddit", "asyncpraw.models.UserSubreddit"]
        ] = None,
        title: Optional[str] = None,
        url: Optional[str] = None,
        **draft_kwargs,
    ):
        data = {
            "body": selftext or url,
            "flair_id": flair_id,
            "flair_text": flair_text,
            "is_public_link": is_public_link,
            "kind": "markdown" if selftext is not None else "link",
            "nsfw": nsfw,
            "original_content": original_content,
            "send_replies": send_replies,
            "spoiler": spoiler,
            "title": title,
        }
        if subreddit:
            if not subreddit._fetched:
                await subreddit.load()
            data.update(
                {
                    "subreddit": subreddit.fullname,
                    "target": "profile"
                    if subreddit.display_name.startswith("u_")
                    else "subreddit",
                }
            )
        data.update(draft_kwargs)
        return data

    def __init__(
        self,
        reddit: "asyncpraw.Reddit",
        id: Optional[str] = None,  # pylint: disable=redefined-builtin
        _data: Dict[str, Any] = None,
    ):
        """Initialize a :class:`.Draft` instance."""
        if (id, _data).count(None) != 1:
            raise TypeError("Exactly one of 'id' or '_data' must be provided.")
        fetched = False
        if id:
            self.id = id
        else:
            if len(_data) > 1:
                if _data["kind"] in ["markdown", "richtext"]:
                    _data["selftext"] = _data.pop("body")
                elif _data["kind"] == "link":
                    _data["url"] = _data.pop("body")
                fetched = True
        super().__init__(reddit, _data=_data, _fetched=fetched)

    def __repr__(self) -> str:
        """Return an object initialization representation of the instance."""
        if self._fetched:
            subreddit = (
                f" subreddit={self.subreddit.display_name!r}" if self.subreddit else ""
            )
            title = f" title={self.title!r}" if self.title else ""
            return f"{self.__class__.__name__}(id={self.id!r}{subreddit}{title})"
        else:
            return f"{self.__class__.__name__}(id={self.id!r})"

    async def _fetch(self):
        async for draft in self._reddit.drafts:
            if draft.id == self.id:
                self.__dict__.update(draft.__dict__)
                self._fetched = True
                return
        raise ClientException(
            f"The currently authenticated user not have a draft with an ID of {self.id}"
        )

    async def delete(self):
        """Delete the :class:`.Draft`.

        Example usage:

        .. code-block:: python

            draft = await reddit.drafts("124862bc-e1e9-11eb-aa4f-e68667a77cbb")
            await draft.delete()

        """
        await self._reddit.delete(API_PATH["draft"], params={"draft_id": self.id})

    async def update(
        self,
        *,
        flair_id: Optional[str] = None,
        flair_text: Optional[str] = None,
        is_public_link: Optional[bool] = None,
        nsfw: Optional[bool] = None,
        original_content: Optional[bool] = None,
        selftext: Optional[str] = None,
        send_replies: Optional[bool] = None,
        spoiler: Optional[bool] = None,
        subreddit: Optional[
            Union[str, "asyncpraw.models.Subreddit", "asyncpraw.models.UserSubreddit"]
        ] = None,
        title: Optional[str] = None,
        url: Optional[str] = None,
        **draft_kwargs,
    ):
        """Update the :class:`.Draft`.

        .. note::

            Only provided values will be updated.

        :param flair_id: The flair template to select.
        :param flair_text: If the template's ``flair_text_editable`` value is ``True``,
            this value will set a custom text. ``flair_id`` is required when
            ``flair_text`` is provided.
        :param is_public_link: Whether to enable public viewing of the draft before it
            is submitted.
        :param nsfw: Whether the draft should be marked NSFW.
        :param original_content: Whether the submission should be marked as original
            content.
        :param selftext: The Markdown formatted content for a text submission draft. Use
            ``None`` to make a title-only submission draft. ``selftext`` can not be
            provided if ``url`` is provided.
        :param send_replies: When ``True``, messages will be sent to the submission
            author when comments are made to the submission.
        :param spoiler: Whether the submission should be marked as a spoiler.
        :param subreddit: The subreddit to create the draft for. This accepts a
            subreddit display name, :class:`.Subreddit` object, or
            :class:`.UserSubreddit` object.
        :param title: The title of the draft.
        :param url: The URL for a ``link`` submission draft. ``url`` can not be provided
            if ``selftext`` is provided.

        Additional keyword arguments can be provided to handle new parameters as Reddit
        introduces them.

        For example, to update the title of a draft do:

        .. code-block:: python

            draft = await reddit.drafts("5f87d55c-e4fb-11eb-8965-6aeb41b0880e")
            await draft.update(title="New title")

        """
        if isinstance(subreddit, str):
            subreddit = await self._reddit.subreddit(subreddit)
        data = await self._prepare_data(
            flair_id=flair_id,
            flair_text=flair_text,
            is_public_link=is_public_link,
            nsfw=nsfw,
            original_content=original_content,
            selftext=selftext,
            send_replies=send_replies,
            spoiler=spoiler,
            subreddit=subreddit,
            title=title,
            url=url,
            **draft_kwargs,
        )
        data["id"] = self.id
        _new_draft = await self._reddit.put(API_PATH["draft"], data=data)
        await _new_draft._fetch()
        self.__dict__.update(_new_draft.__dict__)

    async def submit(
        self,
        *,
        flair_id: Optional[str] = None,
        flair_text: Optional[str] = None,
        nsfw: Optional[bool] = None,
        selftext: Optional[str] = None,
        spoiler: Optional[bool] = None,
        subreddit: Optional[
            Union[str, "asyncpraw.models.Subreddit", "asyncpraw.models.UserSubreddit"]
        ] = None,
        title: Optional[str] = None,
        url: Optional[str] = None,
        **submit_kwargs,
    ) -> "asyncpraw.models.Submission":
        """Submit a draft.

        :param flair_id: The flair template to select (default: ``None``).
        :param flair_text: If the template's ``flair_text_editable`` value is ``True``,
            this value will set a custom text (default: ``None``). ``flair_id`` is
            required when ``flair_text`` is provided.
        :param nsfw: Whether or not the submission should be marked NSFW (default:
            ``None``).
        :param selftext: The Markdown formatted content for a ``text`` submission. Use
            an empty string, ``""``, to make a title-only submission (default:
            ``None``).
        :param spoiler: Whether or not the submission should be marked as a spoiler
            (default: ``None``).
        :param subreddit: The subreddit to submit the draft to. This accepts a subreddit
            display name, :class:`.Subreddit` object, or :class:`.UserSubreddit` object.
        :param title: The title of the submission (default: ``None``).
        :param url: The URL for a ``link`` submission (default: ``None``).

        :returns: A :class:`.Submission` object for the newly created submission.

        .. note::

            Parameters set here will override their respective :class:`.Draft`
            attributes.

        Additional keyword arguments are passed to the :meth:`.Subreddit.submit` method.

        For example, to submit a draft as is:

        .. code-block:: python

            draft = await reddit.drafts("5f87d55c-e4fb-11eb-8965-6aeb41b0880e")
            submission = await draft.submit()

        For example, to submit a draft but use a different title than what is set:

        .. code-block:: python

            draft = reddit.drafts("5f87d55c-e4fb-11eb-8965-6aeb41b0880e")
            submission = draft.submit(title="New Title")

        .. seealso::

            - :meth:`~.Subreddit.submit` to submit url posts and selftexts
            - :meth:`~.Subreddit.submit_gallery`. to submit more than one image in the
              same post
            - :meth:`~.Subreddit.submit_image` to submit images
            - :meth:`~.Subreddit.submit_poll` to submit polls
            - :meth:`~.Subreddit.submit_video` to submit videos and videogifs

        """
        submit_kwargs["draft_id"] = self.id
        if not (self.subreddit or subreddit):
            raise ValueError(
                "'subreddit' must be set on the Draft instance or passed as a keyword"
                " argument."
            )
        for key, attribute in [
            ("flair_id", flair_id),
            ("flair_text", flair_text),
            ("nsfw", nsfw),
            ("selftext", selftext),
            ("spoiler", spoiler),
            ("title", title),
            ("url", url),
        ]:
            value = attribute or getattr(self, key, None)
            if value is not None:
                submit_kwargs[key] = value
        if isinstance(subreddit, str):
            _subreddit = await self._reddit.subreddit(subreddit)
        elif isinstance(subreddit, (Subreddit, UserSubreddit)):
            _subreddit = subreddit
        else:
            _subreddit = self.subreddit
        return await _subreddit.submit(**submit_kwargs)



"""Provide the ReportableMixin class."""
from ....const import API_PATH


class ReportableMixin:
    """Interface for :class:`.RedditBase` classes that can be reported."""

    async def report(self, reason: str):
        """Report this object to the moderators of its subreddit.

        :param reason: The reason for reporting.

        :raises: :class:`.RedditAPIException` if ``reason`` is longer than 100
            characters.

        Example usage:

        .. code-block:: python

            submission = await reddit.submission("5or86n", fetch=False)
            await submission.report("report reason")

            comment = await reddit.comment("dxolpyc", fetch=False)
            await comment.report("report reason")

        """
        await self._reddit.post(
            API_PATH["report"], data={"id": self.fullname, "reason": reason}
        )



"""Test asyncpraw.models.comment_forest."""
import sys

import pytest

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.exceptions import DuplicateReplaceException
from asyncpraw.models import Comment, MoreComments

from .. import IntegrationTest


class TestCommentForest(IntegrationTest):
    def setUp(self):
        super().setUp()
        # Responses do not decode well on travis so manually re-enable gzip.
        self.reddit._core._requestor._http._default_headers["Accept-Encoding"] = "gzip"

    async def test_replace__all(self):
        with self.use_cassette():
            submission = await self.reddit.submission("3hahrw")
            before_count = len(submission.comments.list())
            skipped = await submission.comments.replace_more(limit=None, threshold=0)
            assert len(skipped) == 0
            assert all(isinstance(x, Comment) for x in submission.comments.list())
            assert all(x.submission == submission for x in submission.comments.list())
            assert before_count < len(submission.comments.list())

    async def test_replace__all_large(self):
        with self.use_cassette():
            submission = await self.reddit.submission("n49rw")
            skipped = await submission.comments.replace_more(limit=None, threshold=0)
            assert len(skipped) == 0
            assert all(isinstance(x, Comment) for x in submission.comments.list())
            assert len(submission.comments.list()) > 1000
            assert len(submission.comments.list()) == len(submission._comments_by_id)

    async def test_replace__all_with_comment_limit(self):
        with self.use_cassette():
            submission = await self.reddit.submission("3hahrw")
            submission.comment_limit = 10
            skipped = await submission.comments.replace_more(limit=None, threshold=0)
            assert len(skipped) == 0
            assert len(submission.comments.list()) >= 500

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_replace__all_with_comment_sort(self, _):
        with self.use_cassette():
            submission = await self.reddit.submission("3hahrw", fetch=False)
            submission.comment_sort = "old"
            await submission.load()
            skipped = await submission.comments.replace_more(limit=None, threshold=0)
            assert len(skipped) == 0
            assert len(submission.comments.list()) >= 500

    async def test_replace__skip_at_limit(self):
        with self.use_cassette():
            submission = await self.reddit.submission("3hahrw")
            skipped = await submission.comments.replace_more(limit=1)
            assert len(skipped) == 5

    # def test_replace__skip_below_threshold(self): # FIXME: not currently working; same with praw
    #     with self.use_cassette():
    #         submission = await self.reddit.submission("3hahrw")
    #         before_count = len(submission.comments.list())
    #         skipped = submission.comments.replace_more(limit=16, threshold=5)
    #         assert len(skipped) == 13
    #         assert all(x.count < 5 for x in skipped)
    #         assert all(x.submission == submission for x in skipped)
    #         assert before_count < len(submission.comments.list())

    async def test_replace__skip_all(self):
        with self.use_cassette():
            submission = await self.reddit.submission("3hahrw")
            before_count = len(submission.comments.list())
            skipped = await submission.comments.replace_more(limit=0)
            assert len(skipped) == 6
            assert all(x.submission == submission for x in skipped)
            after_count = len(submission.comments.list())
            assert before_count == after_count + len(skipped)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_replace__on_comment_from_submission(self, _):
        with self.use_cassette():
            submission = await self.reddit.submission("3hahrw")
            types = [type(x) for x in submission.comments.list()]
            assert types.count(Comment) == 527
            assert types.count(MoreComments) == 6
            assert (await submission.comments[0].replies.replace_more()) == []
            types = [type(x) for x in submission.comments.list()]
            assert types.count(Comment) == 531
            assert types.count(MoreComments) == 3

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_replace__on_direct_comment(self, _):
        with self.use_cassette():
            comment = await self.reddit.comment("d8r4im1", fetch=False)
            await comment.refresh()
            assert any(isinstance(x, MoreComments) for x in comment.replies.list())
            await comment.replies.replace_more()
            assert all(isinstance(x, Comment) for x in comment.replies.list())

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_comment_forest_refresh_error(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            submission = await self.async_next(self.reddit.front.top())
            submission.comment_limit = 1
            await submission.load()
            await submission.comments[1].comments()
            with pytest.raises(DuplicateReplaceException):
                await submission.comments.replace_more(limit=1)



import pytest

from asyncpraw.models import Redditor

from ... import UnitTest


class TestRedditor(UnitTest):
    def test_equality(self, reddit):
        redditor1 = Redditor(reddit, _data={"name": "dummy1", "n": 1})
        redditor2 = Redditor(reddit, _data={"name": "Dummy1", "n": 2})
        redditor3 = Redditor(reddit, _data={"name": "dummy3", "n": 2})
        assert redditor1 == redditor1
        assert redditor2 == redditor2
        assert redditor3 == redditor3
        assert redditor1 == redditor2
        assert redditor2 != redditor3
        assert redditor1 != redditor3
        assert "dummy1" == redditor1
        assert redditor2 == "dummy1"

    def test_construct_failure(self, reddit):
        message = "Exactly one of 'name', 'fullname', or '_data' must be provided."
        with pytest.raises(TypeError) as excinfo:
            Redditor(reddit)
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            Redditor(reddit, "dummy", _data={"id": "dummy"})
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            Redditor(reddit, name="dummy", fullname="t2_dummy")
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            Redditor(
                reddit,
                name="dummy",
                fullname="t2_dummy",
                _data={"id": "dummy"},
            )
        assert str(excinfo.value) == message

        with pytest.raises(AssertionError):
            Redditor(reddit, _data=[{"name": "dummy"}])

        with pytest.raises(AssertionError):
            Redditor(reddit, _data={"notname": "dummy"})

        with pytest.raises(ValueError):
            Redditor(reddit, "")
        with pytest.raises(ValueError):
            Redditor(reddit, fullname="")

    def test_fullname(self, reddit):
        redditor = Redditor(reddit, _data={"name": "name", "id": "dummy"})
        assert redditor.fullname == "t2_dummy"

    async def test_guild__min(self, reddit):
        with pytest.raises(TypeError) as excinfo:
            await Redditor(reddit, name="RedditorName").gild(months=0)
        assert str(excinfo.value) == "months must be between 1 and 36"

    async def test_guild__max(self, reddit):
        with pytest.raises(TypeError) as excinfo:
            await Redditor(reddit, name="RedditorName").gild(months=37)
        assert str(excinfo.value) == "months must be between 1 and 36"

    def test_hash(self, reddit):
        redditor1 = Redditor(reddit, _data={"name": "dummy1", "n": 1})
        redditor2 = Redditor(reddit, _data={"name": "Dummy1", "n": 2})
        redditor3 = Redditor(reddit, _data={"name": "dummy3", "n": 2})
        assert hash(redditor1) == hash(redditor1)
        assert hash(redditor2) == hash(redditor2)
        assert hash(redditor3) == hash(redditor3)
        assert hash(redditor1) == hash(redditor2)
        assert hash(redditor2) != hash(redditor3)
        assert hash(redditor1) != hash(redditor3)

    def test_repr(self, reddit):
        redditor = Redditor(reddit, name="RedditorName")
        assert repr(redditor) == "Redditor(name='RedditorName')"

    def test_str(self, reddit):
        redditor = Redditor(reddit, _data={"name": "name", "id": "dummy"})
        assert str(redditor) == "name"


class TestRedditorListings(UnitTest):
    def test__params_not_modified_in_mixed_listing(self, reddit):
        params = {"dummy": "value"}
        redditor = Redditor(reddit, name="spez")
        for listing in ["controversial", "hot", "new", "top"]:
            generator = getattr(redditor, listing)(params=params)
            assert params == {"dummy": "value"}
            assert listing == generator.params["sort"]
            assert "value" == generator.params["dummy"]



from asyncpraw.models import MoreComments

from ... import IntegrationTest


class TestMore(IntegrationTest):
    def setup_method(self, method):
        super().setup_method(method)
        # Responses do not decode well on travis so manually renable gzip.
        self.reddit._core._requestor._http._default_headers["Accept-Encoding"] = "gzip"

    async def test_comments(self):
        data = {
            "count": 9,
            "name": "t1_cu5tt8h",
            "id": "cu5tt8h",
            "parent_id": "t3_3hahrw",
            "children": [
                "cu5tt8h",
                "cu5v9yd",
                "cu5twf5",
                "cu5tkk4",
                "cu5tead",
                "cu5rxpy",
                "cu5oufs",
                "cu5tpek",
                "cu5pbdh",
            ],
        }
        with self.use_cassette():
            more = MoreComments(self.reddit, data)
            more.submission = await self.reddit.submission("3hahrw")
            assert len((await more.comments())) == 7

    async def test_comments__continue_thread_type(self):
        data = {
            "count": 0,
            "name": "t1__",
            "id": "_",
            "parent_id": "t1_cu5v5h7",
            "children": [],
        }
        with self.use_cassette():
            more = MoreComments(self.reddit, data)
            more.submission = await self.reddit.submission("3hahrw")
            assert len((await more.comments())) == 1



from asyncpraw.models import MoreComments

from ... import UnitTest


class TestComment(UnitTest):
    def test_repr(self, reddit):
        more = MoreComments(reddit, {"children": ["a", "b", "c", "d", "e"], "count": 5})
        assert repr(more) == "<MoreComments count=5, children=['a', 'b', 'c', '...']>"

        more = MoreComments(reddit, {"children": ["a", "b", "c", "d"], "count": 4})
        assert repr(more) == "<MoreComments count=4, children=['a', 'b', 'c', 'd']>"

    def test_equality(self, reddit):
        more = MoreComments(reddit, {"children": ["a", "b", "c", "d"], "count": 4})
        more2 = MoreComments(reddit, {"children": ["a", "b", "c", "d"], "count": 4})
        assert more == more2
        assert more != 5



import sys

import pytest

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.exceptions import ClientException, RedditAPIException
from asyncpraw.models import Rule

from ... import IntegrationTest


class TestRule(IntegrationTest):
    async def test_add_rule(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            rule = await subreddit.rules.mod.add(
                short_name="PRAW Test",
                kind="all",
                description="Test by Async PRAW",
                violation_reason="PTest",
            )
        assert rule.short_name == "PRAW Test"
        assert rule.kind == "all"
        assert rule.description == "Test by Async PRAW"
        assert rule.violation_reason == "PTest"

    async def test_add_rule_without_violation_reason(self):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            rule = await subreddit.rules.mod.add(
                short_name="PRAW Test 2", kind="comment"
            )
            assert rule.short_name == "PRAW Test 2"
            assert rule.kind == "comment"
            assert rule.description == ""
            assert rule.violation_reason == "PRAW Test 2"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_delete_rule(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            rules = await self.async_list(subreddit.rules)
            rule = rules[-1]
            await rule.mod.delete()
            assert len(await self.async_list(subreddit.rules)) == (len(rules) - 1)

    async def test_aiter_rules(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            async for rule in subreddit.rules:
                assert isinstance(rule, Rule)

    @pytest.mark.filterwarnings("ignore", category=DeprecationWarning)
    async def test_aiter_call(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            rules = await subreddit.rules()
            assert rules["rules"][0]["short_name"] == "Test post 12"

    async def test_iter_rule_string(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette("TestRule.test_aiter_rules"):
            rule = await subreddit.rules.get_rule("PRAW Test")
            assert isinstance(rule, Rule)
            assert rule.kind

    async def test_iter_rule_invalid(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette("TestRule.test_aiter_rules"):
            with pytest.raises(ClientException) as excinfo:
                await subreddit.rules.get_rule("fake rule")
            assert (
                excinfo.value.args[0]
                == f"Subreddit {subreddit} does not have the rule fake rule"
            )

    async def test_iter_rule_int(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette("TestRule.test_aiter_rules"):
            rule = await subreddit.rules.get_rule(0)
            assert isinstance(rule, Rule)

    async def test_iter_rule_negative_int(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette("TestRule.test_aiter_rules"):
            rule = await subreddit.rules.get_rule(-1)
            assert isinstance(rule, Rule)

    async def test_iter_rule_slice(self):
        with self.use_cassette("TestRule.test_aiter_rules"):
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            rules = await subreddit.rules.get_rule(slice(-3, None))
            assert len(rules) == 3
            for rule in rules:
                assert isinstance(rule, Rule)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_reorder_rules(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            rule_list = await self.async_list(subreddit.rules)
            reordered = rule_list[2:3] + rule_list[0:2] + rule_list[3:]
            rule_info = {rule.short_name: rule for rule in rule_list}
            await subreddit.rules.mod.reorder(reordered)
            new_rules = await self.async_list(subreddit.rules)
            assert new_rules != rule_list
            for rule in new_rules:
                assert rule_info[rule.short_name] == rule

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_reorder_rules_double(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            rule_list = await self.async_list(subreddit.rules)
            with pytest.raises(RedditAPIException):
                await subreddit.rules.mod.reorder(rule_list + rule_list[0:1])

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_reorder_rules_empty(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            with pytest.raises(RedditAPIException):
                await subreddit.rules.mod.reorder([])

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_reorder_rules_no_reorder(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            rule_list = await self.async_list(subreddit.rules)
            new_list = await subreddit.rules.mod.reorder(rule_list)
            assert new_list == rule_list

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_reorder_rules_omit(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            rule_list = await self.async_list(subreddit.rules)
            with pytest.raises(RedditAPIException):
                await subreddit.rules.mod.reorder(rule_list[:-1])

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update_rule(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            rule_list = await self.async_list(subreddit.rules)
            rule = rule_list[0]
            rule2 = await rule.mod.update(
                description="Updated rule",
                kind="link",
                violation_reason="PUpdate",
            )
            assert rule.description != rule2.description
            assert rule2.description == "Updated rule"
            assert rule.kind != rule2.kind
            assert rule2.kind == "link"
            assert rule.violation_reason != rule2.violation_reason
            assert rule2.violation_reason == "PUpdate"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update_rule_short_name(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            rule = await subreddit.rules.get_rule("Test Rule")
            rule2 = await rule.mod.update(
                short_name="PRAW Update",
                description="Updated rule",
                kind="comment",
                violation_reason="PUpdate",
            )
            assert rule != rule2
            assert rule2.short_name == "PRAW Update"
            assert rule.description != rule2.description
            assert rule2.description == "Updated rule"
            assert rule.kind != rule2.kind
            assert rule2.kind == "comment"
            assert rule.violation_reason != rule2.violation_reason
            assert rule2.violation_reason == "PUpdate"
            async for new_rule in subreddit.rules:
                assert new_rule.short_name != rule.short_name

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update_rule_no_params(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            rule = await subreddit.rules.get_rule("Test Rule")
            rule2 = await rule.mod.update()
            for attr in (
                "created_utc",
                "description",
                "kind",
                "priority",
                "short_name",
                "subreddit",
                "violation_reason",
            ):
                assert getattr(rule, attr) == getattr(rule2, attr)



#!/usr/bin/env python3
"""This example demonstrates using the sqlite token manager for refresh tokens.

In order to run this program, you will first need to obtain one or more valid refresh
tokens. You can use the ``obtain_refresh_token.py`` example to help.

In this example, refresh tokens will be saved into a file ``tokens.sqlite3`` relative to
your current working directory. If your current working directory is under version
control it is strongly encouraged you add ``tokens.sqlite3`` to the version control ignore
list.

This example differs primarily from ``use_file_token_manager.py`` due to the fact that a
shared SQLite3 database can manage many ``refresh_tokens``. While each instance of
Reddit still needs to have 1-to-1 mapping to a token manager, multiple Reddit instances
can concurrently share access to the same SQLite3 database; the same cannot be done with
the FileTokenManager.

Usage:

    export praw_client_id=hIkgfjAZpV9u4A
    export praw_client_secret=jHXu3riMfnn2vEsUDTHMgSNjFWA
    python use_sqlite_token_manager.py test

"""
import asyncio
import os
import sys

import asyncpraw
from asyncpraw.util.token_manager import SQLiteTokenManager

DATABASE_PATH = "tokens.sqlite3"


async def main():
    if "praw_client_id" not in os.environ:
        sys.stderr.write("Environment variable 'praw_client_id' must be defined\n")
        return 1
    if "praw_client_secret" not in os.environ:
        sys.stderr.write("Environment variable 'praw_client_secret' must be defined\n")
        return 1
    if len(sys.argv) != 2:
        sys.stderr.write(
            "'KEY' must be provided.\n\nUsage: python3 use_sqlite_token_manager.py"
            " TOKEN_KEY\n"
        )
        return 1

    refresh_token_manager = SQLiteTokenManager(database=DATABASE_PATH, key=sys.argv[1])
    reddit = asyncpraw.Reddit(
        token_manager=refresh_token_manager,
        user_agent="sqlite_token_manager/v0 by u/bboe",
    )

    if not await refresh_token_manager.is_registered():
        refresh_token = input("Enter initial refresh token: ").strip()
        await refresh_token_manager.register(refresh_token)

    scopes = await reddit.auth.scopes()
    if scopes == {"*"}:
        print(f"{await reddit.user.me()} is authenticated with all scopes")
    elif "identity" in scopes:
        print(
            f"{await reddit.user.me()} is authenticated with the following scopes:"
            f" {scopes}"
        )
    else:
        print(f"You are authenticated with the following scopes: {scopes}")
    await refresh_token_manager.close()
    await reddit.close()


if __name__ == "__main__":
    sys.exit(asyncio.run(main()))



"""Test classes from collections.py."""
import pytest

from asyncpraw.models import Collection
from asyncpraw.models.reddit.collections import SubredditCollections

from ... import UnitTest


class TestCollection(UnitTest):
    def test_eq(self, reddit):
        uuid = "fake_uuid"
        permalink = f"https://reddit.com/r/subreddit/collection/{uuid}"
        collection1 = Collection(reddit, collection_id=uuid)
        collection2 = Collection(reddit, permalink=permalink)
        assert collection1 == collection2
        assert collection2 == collection2
        assert collection1 == collection1
        assert uuid == collection1
        assert uuid == collection2

    def test_init(self, reddit):
        uuid = "fake_uuid"
        assert uuid == Collection(reddit, collection_id=uuid).collection_id
        permalink = f"https://reddit.com/r/subreddit/collection/{uuid}"
        assert uuid == Collection(reddit, permalink=permalink).collection_id

    def test_init_bad(self, reddit):
        with pytest.raises(TypeError):
            Collection(reddit)
        with pytest.raises(TypeError):
            Collection(reddit, _data=dict(), collection_id="")
        with pytest.raises(TypeError):
            Collection(reddit, collection_id="fake_uuid", permalink="")
        with pytest.raises(TypeError):
            Collection(
                None,
                _data={"collection_id": "fake_uuid"},
                collection_id="fake_uuid",
                permalink="https://reddit.com/r/sub/collection/fake_uuid",
            )
        with pytest.raises(ValueError):
            Collection(reddit, collection_id="")
        with pytest.raises(ValueError):
            Collection(reddit, permalink="")

    def test_repr(self, reddit):
        collection = Collection(reddit, collection_id="fake_uuid")
        assert "Collection(collection_id='fake_uuid')" == repr(collection)

    def test_str(self, reddit):
        collection = Collection(reddit, collection_id="fake_uuid")
        assert "fake_uuid" == str(collection)

    def test_neq(self, reddit):
        collection1 = Collection(reddit, collection_id="1")
        collection2 = Collection(reddit, collection_id="2")
        assert collection1 != collection2
        assert "1" != collection2
        assert "2" != collection1


class TestSubredditCollections(UnitTest):
    async def test_call(self, reddit):
        collections = SubredditCollections(
            reddit, await reddit.subreddit("placeholder")
        )
        with pytest.raises(TypeError):
            await collections()
        with pytest.raises(TypeError):
            await collections("a uuid", "a permalink")



"""Provide the EditableMixin class."""
from typing import TYPE_CHECKING, Union

from ....const import API_PATH

if TYPE_CHECKING:  # pragma: no cover
    import asyncpraw


class EditableMixin:
    """Interface for classes that can be edited and deleted."""

    async def delete(self):
        """Delete the object.

        Example usage:

        .. code-block:: python

            comment = await reddit.comment("dkk4qjd", fetch=False)
            await comment.delete()

            submission = await reddit.submission("8dmv8z", fetch=False)
            await submission.delete()

        """
        await self._reddit.post(API_PATH["del"], data={"id": self.fullname})

    async def edit(
        self, body: str
    ) -> Union["asyncpraw.models.Comment", "asyncpraw.models.Submission"]:
        """Replace the body of the object with ``body``.

        :param body: The Markdown formatted content for the updated object.

        :returns: The current instance after updating its attributes.

        Example usage:

        .. code-block:: python

            comment = await reddit.comment("dkk4qjd")

            # construct the text of an edited comment
            # by appending to the old body:
            edited_body = comment.body + "Edit: thanks for the gold!"
            await comment.edit(edited_body)

        """
        data = {
            "text": body,
            "thing_id": self.fullname,
            "validate_on_submit": self._reddit.validate_on_submit,
        }
        response = await self._reddit.post(API_PATH["edit"], data=data)
        updated = response[0]
        for attribute in [
            "_fetched",
            "_reddit",
            "_submission",
            "replies",
            "subreddit",
        ]:
            if attribute in updated.__dict__:
                delattr(updated, attribute)
        self.__dict__.update(updated.__dict__)
        return self
